# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.27.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from gentrace import schemas  # noqa: F401


class RunV2(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "stepRuns",
            "id",
        }
        
        class properties:
            id = schemas.UUIDSchema
            
            
            class stepRuns(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    
                    
                    class items(
                        schemas.DictSchema
                    ):
                    
                    
                        class MetaOapg:
                            required = {
                                "outputs",
                                "invocation",
                                "cost",
                                "modelParams",
                                "inputs",
                                "stepId",
                                "startTime",
                                "endTime",
                                "providerName",
                                "stepRunId",
                                "elapsedTime",
                                "order",
                            }
                            
                            class properties:
                                stepRunId = schemas.UUIDSchema
                                stepId = schemas.UUIDSchema
                                providerName = schemas.StrSchema
                                invocation = schemas.StrSchema
                                order = schemas.IntSchema
                                modelParams = schemas.DictSchema
                                inputs = schemas.DictSchema
                                outputs = schemas.DictSchema
                                elapsedTime = schemas.IntSchema
                                startTime = schemas.Float32Schema
                                endTime = schemas.Float32Schema
                                cost = schemas.NumberSchema
                                
                                
                                class metadata(
                                    schemas.DictBase,
                                    schemas.NoneBase,
                                    schemas.Schema,
                                    schemas.NoneFrozenDictMixin
                                ):
                                
                                
                                    class MetaOapg:
                                        
                                        @staticmethod
                                        def additional_properties() -> typing.Type['MetadataValueObject']:
                                            return MetadataValueObject
                                
                                    
                                    def __getitem__(self, name: typing.Union[str, ]) -> 'MetadataValueObject':
                                        # dict_instance[name] accessor
                                        return super().__getitem__(name)
                                    
                                    def get_item_oapg(self, name: typing.Union[str, ]) -> 'MetadataValueObject':
                                        return super().get_item_oapg(name)
                                
                                    def __new__(
                                        cls,
                                        *_args: typing.Union[dict, frozendict.frozendict, None, ],
                                        _configuration: typing.Optional[schemas.Configuration] = None,
                                        **kwargs: 'MetadataValueObject',
                                    ) -> 'metadata':
                                        return super().__new__(
                                            cls,
                                            *_args,
                                            _configuration=_configuration,
                                            **kwargs,
                                        )
                                
                                
                                class error(
                                    schemas.StrBase,
                                    schemas.NoneBase,
                                    schemas.Schema,
                                    schemas.NoneStrMixin
                                ):
                                
                                
                                    def __new__(
                                        cls,
                                        *_args: typing.Union[None, str, ],
                                        _configuration: typing.Optional[schemas.Configuration] = None,
                                    ) -> 'error':
                                        return super().__new__(
                                            cls,
                                            *_args,
                                            _configuration=_configuration,
                                        )
                                __annotations__ = {
                                    "stepRunId": stepRunId,
                                    "stepId": stepId,
                                    "providerName": providerName,
                                    "invocation": invocation,
                                    "order": order,
                                    "modelParams": modelParams,
                                    "inputs": inputs,
                                    "outputs": outputs,
                                    "elapsedTime": elapsedTime,
                                    "startTime": startTime,
                                    "endTime": endTime,
                                    "cost": cost,
                                    "metadata": metadata,
                                    "error": error,
                                }
                        
                        outputs: MetaOapg.properties.outputs
                        invocation: MetaOapg.properties.invocation
                        cost: MetaOapg.properties.cost
                        modelParams: MetaOapg.properties.modelParams
                        inputs: MetaOapg.properties.inputs
                        stepId: MetaOapg.properties.stepId
                        startTime: MetaOapg.properties.startTime
                        endTime: MetaOapg.properties.endTime
                        providerName: MetaOapg.properties.providerName
                        stepRunId: MetaOapg.properties.stepRunId
                        elapsedTime: MetaOapg.properties.elapsedTime
                        order: MetaOapg.properties.order
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["stepRunId"]) -> MetaOapg.properties.stepRunId: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["stepId"]) -> MetaOapg.properties.stepId: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["providerName"]) -> MetaOapg.properties.providerName: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["invocation"]) -> MetaOapg.properties.invocation: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["order"]) -> MetaOapg.properties.order: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["modelParams"]) -> MetaOapg.properties.modelParams: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["inputs"]) -> MetaOapg.properties.inputs: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["outputs"]) -> MetaOapg.properties.outputs: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["elapsedTime"]) -> MetaOapg.properties.elapsedTime: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["startTime"]) -> MetaOapg.properties.startTime: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["endTime"]) -> MetaOapg.properties.endTime: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["cost"]) -> MetaOapg.properties.cost: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["metadata"]) -> MetaOapg.properties.metadata: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["error"]) -> MetaOapg.properties.error: ...
                        
                        @typing.overload
                        def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
                        
                        def __getitem__(self, name: typing.Union[typing_extensions.Literal["stepRunId", "stepId", "providerName", "invocation", "order", "modelParams", "inputs", "outputs", "elapsedTime", "startTime", "endTime", "cost", "metadata", "error", ], str]):
                            # dict_instance[name] accessor
                            return super().__getitem__(name)
                        
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["stepRunId"]) -> MetaOapg.properties.stepRunId: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["stepId"]) -> MetaOapg.properties.stepId: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["providerName"]) -> MetaOapg.properties.providerName: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["invocation"]) -> MetaOapg.properties.invocation: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["order"]) -> MetaOapg.properties.order: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["modelParams"]) -> MetaOapg.properties.modelParams: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["inputs"]) -> MetaOapg.properties.inputs: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["outputs"]) -> MetaOapg.properties.outputs: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["elapsedTime"]) -> MetaOapg.properties.elapsedTime: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["startTime"]) -> MetaOapg.properties.startTime: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["endTime"]) -> MetaOapg.properties.endTime: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["cost"]) -> MetaOapg.properties.cost: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["metadata"]) -> typing.Union[MetaOapg.properties.metadata, schemas.Unset]: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["error"]) -> typing.Union[MetaOapg.properties.error, schemas.Unset]: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
                        
                        def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["stepRunId", "stepId", "providerName", "invocation", "order", "modelParams", "inputs", "outputs", "elapsedTime", "startTime", "endTime", "cost", "metadata", "error", ], str]):
                            return super().get_item_oapg(name)
                        
                    
                        def __new__(
                            cls,
                            *_args: typing.Union[dict, frozendict.frozendict, ],
                            outputs: typing.Union[MetaOapg.properties.outputs, dict, frozendict.frozendict, ],
                            invocation: typing.Union[MetaOapg.properties.invocation, str, ],
                            cost: typing.Union[MetaOapg.properties.cost, decimal.Decimal, int, float, ],
                            modelParams: typing.Union[MetaOapg.properties.modelParams, dict, frozendict.frozendict, ],
                            inputs: typing.Union[MetaOapg.properties.inputs, dict, frozendict.frozendict, ],
                            stepId: typing.Union[MetaOapg.properties.stepId, str, uuid.UUID, ],
                            startTime: typing.Union[MetaOapg.properties.startTime, decimal.Decimal, int, float, ],
                            endTime: typing.Union[MetaOapg.properties.endTime, decimal.Decimal, int, float, ],
                            providerName: typing.Union[MetaOapg.properties.providerName, str, ],
                            stepRunId: typing.Union[MetaOapg.properties.stepRunId, str, uuid.UUID, ],
                            elapsedTime: typing.Union[MetaOapg.properties.elapsedTime, decimal.Decimal, int, ],
                            order: typing.Union[MetaOapg.properties.order, decimal.Decimal, int, ],
                            metadata: typing.Union[MetaOapg.properties.metadata, dict, frozendict.frozendict, None, schemas.Unset] = schemas.unset,
                            error: typing.Union[MetaOapg.properties.error, None, str, schemas.Unset] = schemas.unset,
                            _configuration: typing.Optional[schemas.Configuration] = None,
                            **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                        ) -> 'items':
                            return super().__new__(
                                cls,
                                *_args,
                                outputs=outputs,
                                invocation=invocation,
                                cost=cost,
                                modelParams=modelParams,
                                inputs=inputs,
                                stepId=stepId,
                                startTime=startTime,
                                endTime=endTime,
                                providerName=providerName,
                                stepRunId=stepRunId,
                                elapsedTime=elapsedTime,
                                order=order,
                                metadata=metadata,
                                error=error,
                                _configuration=_configuration,
                                **kwargs,
                            )
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[MetaOapg.items, dict, frozendict.frozendict, ]], typing.List[typing.Union[MetaOapg.items, dict, frozendict.frozendict, ]]],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'stepRuns':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)
            __annotations__ = {
                "id": id,
                "stepRuns": stepRuns,
            }
    
    stepRuns: MetaOapg.properties.stepRuns
    id: MetaOapg.properties.id
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["stepRuns"]) -> MetaOapg.properties.stepRuns: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["id", "stepRuns", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["stepRuns"]) -> MetaOapg.properties.stepRuns: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["id", "stepRuns", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        stepRuns: typing.Union[MetaOapg.properties.stepRuns, list, tuple, ],
        id: typing.Union[MetaOapg.properties.id, str, uuid.UUID, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'RunV2':
        return super().__new__(
            cls,
            *_args,
            stepRuns=stepRuns,
            id=id,
            _configuration=_configuration,
            **kwargs,
        )

from gentrace.model.metadata_value_object import MetadataValueObject



from gentrace.providers.vectorstores.pinecone import (
    PineconeDeleteStepRun,
    PineconeFetchStepRun,
    PineconePipelineHandler,
    PineconeQueryStepRun,
    PineconeUpdateStepRun,
    PineconeUpsertStepRun,
)



import time
import uuid
from typing import Any, Dict, List, Optional, Tuple, Union

import pinecone
from pinecone import (
    FetchResponse,
    QueryResponse,
    QueryVector,
    SparseValues,
    UpsertResponse,
    Vector,
)
from pinecone.config import init

from gentrace.configuration import Configuration as GentraceConfiguration
from gentrace.providers.context import Context
from gentrace.providers.init import GENTRACE_CONFIG_STATE
from gentrace.providers.pipeline import Pipeline
from gentrace.providers.pipeline_run import PipelineRun
from gentrace.providers.step_run import StepRun
from gentrace.providers.utils import to_date_string


class ModifiedIndex(pinecone.Index):
    def __init__(
            self,
            index_name: str,
            pool_threads=1,
            *args,
            **kwargs,
    ):
        super().__init__(index_name, pool_threads, *args, **kwargs)

    def manage_step_run(
            self,
            step_run,
            pipeline_slug,
            context: Optional[Context] = None,
    ):
        pipeline_run_id = None
        is_self_contained = not hasattr(self, "pipeline_run") and pipeline_slug
        if is_self_contained:
            pipeline_run_id = str(uuid.uuid4())

        pipeline_run = self.pipeline_run if hasattr(self, "pipeline_run") else None

        if is_self_contained:
            from gentrace import api_key, host

            if api_key:
                gentrace_config = GentraceConfiguration(host=host)
                gentrace_config.access_token = api_key
            else:
                gentrace_config = GENTRACE_CONFIG_STATE["global_gentrace_config"]

            pipeline = Pipeline(
                slug=pipeline_slug,
                api_key=gentrace_config.access_token,
                host=gentrace_config.host,
            )

            pipeline_run = PipelineRun(
                pipeline=pipeline, id=pipeline_run_id, context=context
            )

        if pipeline_run:
            pipeline_run.add_step_run(step_run)

            if is_self_contained:
                pipeline_run.submit()

        return pipeline_run_id

    def fetch(
            self, ids: List[str], namespace: Optional[str] = None, **kwargs
    ) -> FetchResponse:
        start_time = time.time()

        # @deprecated: pipeline_id is deprecated, use pipeline_slug instead
        pipeline_id = kwargs.pop("pipeline_id", None)
        pipeline_slug = kwargs.pop("pipeline_slug", None)

        context = kwargs.pop("gentrace", {})

        effective_pipeline_slug = pipeline_slug or pipeline_id

        response = super().fetch(ids, namespace, **kwargs)
        end_time = time.time()
        elapsed_time = int((end_time - start_time) * 1000)

        pipeline_run_id = self.manage_step_run(
            PineconeFetchStepRun(
                elapsed_time,
                to_date_string(start_time),
                to_date_string(end_time),
                {"ids": ids, "namespace": namespace},
                response.to_dict(),
                context,
            ),
            effective_pipeline_slug,
            context,
        )

        if pipeline_run_id:
            response["pipelineRunId"] = pipeline_run_id

        return response

    def update(
            self,
            id: str,
            values: Optional[List[float]] = None,
            set_metadata: Optional[
                Dict[str, Union[str, float, int, bool, List[int], List[float], List[str]]]
            ] = None,
            namespace: Optional[str] = None,
            sparse_values: Optional[
                Union[SparseValues, Dict[str, Union[List[float], List[int]]]]
            ] = None,
            **kwargs,
    ) -> Dict[str, any]:
        start_time = time.time()
        # @deprecated: pipeline_id is deprecated, use pipeline_slug instead
        pipeline_id = kwargs.pop("pipeline_id", None)
        pipeline_slug = kwargs.pop("pipeline_slug", None)

        context = kwargs.pop("gentrace", {})

        effective_pipeline_slug = pipeline_slug or pipeline_id
        response = super().update(
            values, set_metadata, namespace, sparse_values, **kwargs
        )
        end_time = time.time()
        elapsed_time = int((end_time - start_time) * 1000)

        pipeline_run_id = self.manage_step_run(
            PineconeUpdateStepRun(
                elapsed_time,
                to_date_string(start_time),
                to_date_string(end_time),
                {
                    "id": id,
                    "values": values,
                    "set_metadata": set_metadata,
                    "namespace": namespace,
                    "sparse_values": sparse_values,
                },
                response,
                context,
            ),
            effective_pipeline_slug,
            context,
        )

        if pipeline_run_id:
            response["pipelineRunId"] = pipeline_run_id

        return response

    def query(
            self,
            vector: Optional[List[float]] = None,
            id: Optional[str] = None,
            queries: Optional[Union[List[QueryVector], List[Tuple]]] = None,
            top_k: Optional[int] = None,
            namespace: Optional[str] = None,
            filter: Optional[Dict[str, Union[str, float, int, bool, List, dict]]] = None,
            include_values: Optional[bool] = None,
            include_metadata: Optional[bool] = None,
            sparse_vector: Optional[
                Union[SparseValues, Dict[str, Union[List[float], List[int]]]]
            ] = None,
            **kwargs,
    ) -> QueryResponse:
        # @deprecated: pipeline_id is deprecated, use pipeline_slug instead
        pipeline_id = kwargs.pop("pipeline_id", None)
        pipeline_slug = kwargs.pop("pipeline_slug", None)
        context = kwargs.pop("gentrace", {})
        effective_pipeline_slug = pipeline_slug or pipeline_id
        bound_query = super().query
        start_time = time.time()
        response = bound_query(
            vector,
            id,
            queries,
            top_k,
            namespace,
            filter,
            include_values,
            include_metadata,
            sparse_vector,
            **kwargs,
        )
        end_time = time.time()
        elapsed_time = int((end_time - start_time) * 1000)
        inputs = {
            "vector": vector,
            "id": id,
            "queries": queries,
            "namespace": namespace,
            "include_values": include_values,
            "include_metadata": include_metadata,
            "sparse_vector": sparse_vector,
        }
        model_params = {"top_k": top_k, "filter": filter}

        pipeline_run_id = self.manage_step_run(
            PineconeQueryStepRun(
                elapsed_time,
                to_date_string(start_time),
                to_date_string(end_time),
                {**inputs},
                {**model_params},
                response.to_dict(),
                context,
            ),
            effective_pipeline_slug,
            context,
        )

        if pipeline_run_id:
            response["pipelineRunId"] = pipeline_run_id

        return response

    def upsert(
            self,
            vectors: Union[List[Vector], List[tuple], List[dict]],
            namespace: Optional[str] = None,
            batch_size: Optional[int] = None,
            show_progress: bool = True,
            **kwargs,
    ) -> UpsertResponse:
        # @deprecated: pipeline_id is deprecated, use pipeline_slug instead
        pipeline_id = kwargs.pop("pipeline_id", None)
        pipeline_slug = kwargs.pop("pipeline_slug", None)
        context = kwargs.pop("gentrace", {})
        effective_pipeline_slug = pipeline_slug or pipeline_id

        start_time = time.time()
        response = super().upsert(
            vectors, namespace, batch_size, show_progress, **kwargs
        )
        end_time = time.time()
        elapsed_time = int((end_time - start_time) * 1000)

        pipeline_run_id = self.manage_step_run(
            PineconeUpsertStepRun(
                elapsed_time,
                to_date_string(start_time),
                to_date_string(end_time),
                {
                    "vectors": vectors,
                    "namespace": namespace,
                    "batch_size": batch_size,
                    "show_progress": show_progress,
                },
                response.to_dict(),
                context,
            ),
            effective_pipeline_slug,
            context,
        )

        if pipeline_run_id:
            response["pipelineRunId"] = pipeline_run_id

        return response

    def delete(
            self,
            ids: Optional[List[str]] = None,
            delete_all: Optional[bool] = None,
            namespace: Optional[str] = None,
            filter: Optional[Dict[str, Union[str, float, int, bool, List, dict]]] = None,
            **kwargs,
    ) -> Dict[str, Any]:
        start_time = time.time()
        response = super().delete(ids, delete_all, namespace, filter, **kwargs)
        end_time = time.time()
        elapsed_time = int((end_time - start_time) * 1000)

        # @deprecated: pipeline_id is deprecated, use pipeline_slug instead
        pipeline_id = kwargs.pop("pipeline_id", None)
        pipeline_slug = kwargs.pop("pipeline_slug", None)
        context = kwargs.pop("gentrace", {})

        effective_pipeline_slug = pipeline_slug or pipeline_id

        pipeline_run_id = self.manage_step_run(
            PineconeDeleteStepRun(
                elapsed_time,
                to_date_string(start_time),
                to_date_string(end_time),
                {
                    "ids": ids,
                    "delete_all": delete_all,
                    "namespace": namespace,
                    "filter": filter,
                },
                response,
                context,
            ),
            effective_pipeline_slug,
            context,
        )

        if pipeline_run_id:
            response["pipelineRunId"] = pipeline_run_id

        return response


def Index(index_name: str, pool_threads=1, *args, **kwargs):
    modified_index = ModifiedIndex(index_name, pool_threads, *args, **kwargs)
    return modified_index


class PineconePipelineHandler:
    pipeline: Optional[Pipeline] = None
    pipeline_run: Optional[PipelineRun] = None

    def __init__(self, pipeline=None):
        self.pipeline: Optional[Pipeline] = pipeline

    def __getattr__(self, name):
        import pinecone

        return getattr(pinecone, name)

    def set_pipeline_run(self, pipeline_run):
        self.pipeline_run: Optional[PipelineRun] = pipeline_run

    def init(self, *args, **kwargs):
        init(*args, **kwargs)

    def Index(self, *args, **kwargs):
        modified_index = self.ModifiedIndexContained(self.pipeline_run, *args, **kwargs)
        return modified_index

    class ModifiedIndexContained(ModifiedIndex):
        def __init__(self, pipeline_run, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.pipeline_run = pipeline_run


# Assign exported member function to PineconePipelineHandler
for exported_member in dir(pinecone):
    if (
            not exported_member.startswith("__")
            and exported_member != "Index"
            and exported_member != "init"
    ):
        retrieved_member = getattr(pinecone, exported_member)
        setattr(
            PineconePipelineHandler,
            exported_member,
            staticmethod(retrieved_member)
            if callable(retrieved_member)
            else retrieved_member,
        )


def annotate_pinecone_module():
    # Directly modify the pinecone module by changing the Index class that's used.
    pinecone.Index = Index
    pass


class PineconeFetchStepRun(StepRun):
    def __init__(
            self,
            elapsed_time: int,
            start_time: str,
            end_time: str,
            inputs: dict,
            response: dict,
            context: Optional[Context] = None,
            error: Optional[str] = None,
    ):
        super().__init__(
            "pinecone",
            "pinecone_indexFetch",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            {},
            response,
            context,
            error,
        )
        self.response = response


class PineconeQueryStepRun(StepRun):
    def __init__(
            self,
            elapsed_time: int,
            start_time: str,
            end_time: str,
            inputs: dict,
            model_params: dict,
            response: dict,
            context: Optional[Context] = None,
            error: Optional[str] = None,
    ):
        super().__init__(
            "pinecone",
            "pinecone_indexQuery",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
            error,
        )
        self.response = response


class PineconeUpdateStepRun(StepRun):
    def __init__(
            self,
            elapsed_time: int,
            start_time: str,
            end_time: str,
            inputs: dict,
            response: dict,
            context: Optional[Context] = None,
            error: Optional[str] = None,
    ):
        super().__init__(
            "pinecone",
            "pinecone_indexUpdate",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            {},
            response,
            context,
            error,
        )
        self.response = response


class PineconeUpsertStepRun(StepRun):
    def __init__(
            self,
            elapsed_time: int,
            start_time: str,
            end_time: str,
            inputs: dict,
            response: dict,
            context: Optional[Context] = None,
            error: Optional[str] = None,
    ):
        super().__init__(
            "pinecone",
            "pinecone_indexUpsert",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            {},
            response,
            context,
            error,
        )
        self.response = response


class PineconeDeleteStepRun(StepRun):
    def __init__(
            self,
            elapsed_time: int,
            start_time: str,
            end_time: str,
            inputs: dict,
            response: dict,
            context: Optional[Context] = None,
            error: Optional[str] = None,
    ):
        super().__init__(
            "pinecone",
            "pinecone_indexDelete1",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            {},
            response,
            context,
            error,
        )
        self.response = response



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.27.0
    Generated by: https://openapi-generator.tech
"""

from gentrace.paths.v2_datasets.get import V2DatasetsGet
from gentrace.paths.v2_datasets_id.get import V2DatasetsIdGet
from gentrace.paths.v2_datasets_id.post import V2DatasetsIdPost
from gentrace.paths.v2_datasets.post import V2DatasetsPost
from gentrace.paths.v2_evaluations_bulk.post import V2EvaluationsBulkPost
from gentrace.paths.v2_evaluations.get import V2EvaluationsGet
from gentrace.paths.v2_evaluators.get import V2EvaluatorsGet
from gentrace.paths.v2_feedback_id.get import V2FeedbackIdGet
from gentrace.paths.v2_feedback_id.patch import V2FeedbackIdPatch
from gentrace.paths.v2_feedback.post import V2FeedbackPost
from gentrace.paths.v2_folders.get import V2FoldersGet
from gentrace.paths.v2_folders_id.get import V2FoldersIdGet
from gentrace.paths.v2_pipelines.get import V2PipelinesGet
from gentrace.paths.v2_runs_id.get import V2RunsIdGet
from gentrace.paths.v2_test_cases.get import V2TestCasesGet
from gentrace.paths.v2_test_cases_id.delete import V2TestCasesIdDelete
from gentrace.paths.v2_test_cases_id.get import V2TestCasesIdGet
from gentrace.paths.v2_test_results.get import V2TestResultsGet


class V2Api(
    V2DatasetsGet,
    V2DatasetsIdGet,
    V2DatasetsIdPost,
    V2DatasetsPost,
    V2EvaluationsBulkPost,
    V2EvaluationsGet,
    V2EvaluatorsGet,
    V2FeedbackIdGet,
    V2FeedbackIdPatch,
    V2FeedbackPost,
    V2FoldersGet,
    V2FoldersIdGet,
    V2PipelinesGet,
    V2RunsIdGet,
    V2TestCasesGet,
    V2TestCasesIdDelete,
    V2TestCasesIdGet,
    V2TestResultsGet,
):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    pass



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.27.0
    Generated by: https://openapi-generator.tech
"""

from gentrace.paths.v3_evaluations.get import V3EvaluationsGet


class V3Api(
    V3EvaluationsGet,
):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    pass



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.27.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from gentrace import schemas  # noqa: F401


class RunRequest(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "stepRuns",
            "id",
        }
        
        class properties:
            id = schemas.UUIDSchema
            
            
            class stepRuns(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    
                    @staticmethod
                    def items() -> typing.Type['StepRun']:
                        return StepRun
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['StepRun'], typing.List['StepRun']],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'stepRuns':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> 'StepRun':
                    return super().__getitem__(i)
            
            
            class collectionMethod(
                schemas.EnumBase,
                schemas.StrSchema
            ):
            
            
                class MetaOapg:
                    enum_value_to_name = {
                        "manual": "MANUAL",
                        "runner": "RUNNER",
                    }
                
                @schemas.classproperty
                def MANUAL(cls):
                    return cls("manual")
                
                @schemas.classproperty
                def RUNNER(cls):
                    return cls("runner")
            slug = schemas.StrSchema
            
            
            class previousRunId(
                schemas.UUIDBase,
                schemas.StrBase,
                schemas.NoneBase,
                schemas.Schema,
                schemas.NoneStrMixin
            ):
            
            
                class MetaOapg:
                    format = 'uuid'
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[None, str, uuid.UUID, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'previousRunId':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                    )
            
            
            class metadata(
                schemas.DictBase,
                schemas.NoneBase,
                schemas.Schema,
                schemas.NoneFrozenDictMixin
            ):
            
            
                class MetaOapg:
                    
                    @staticmethod
                    def additional_properties() -> typing.Type['MetadataValueObject']:
                        return MetadataValueObject
            
                
                def __getitem__(self, name: typing.Union[str, ]) -> 'MetadataValueObject':
                    # dict_instance[name] accessor
                    return super().__getitem__(name)
                
                def get_item_oapg(self, name: typing.Union[str, ]) -> 'MetadataValueObject':
                    return super().get_item_oapg(name)
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, None, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: 'MetadataValueObject',
                ) -> 'metadata':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class error(
                schemas.StrBase,
                schemas.NoneBase,
                schemas.Schema,
                schemas.NoneStrMixin
            ):
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[None, str, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'error':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                    )
            __annotations__ = {
                "id": id,
                "stepRuns": stepRuns,
                "collectionMethod": collectionMethod,
                "slug": slug,
                "previousRunId": previousRunId,
                "metadata": metadata,
                "error": error,
            }
    
    stepRuns: MetaOapg.properties.stepRuns
    id: MetaOapg.properties.id
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["stepRuns"]) -> MetaOapg.properties.stepRuns: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["collectionMethod"]) -> MetaOapg.properties.collectionMethod: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["slug"]) -> MetaOapg.properties.slug: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["previousRunId"]) -> MetaOapg.properties.previousRunId: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["metadata"]) -> MetaOapg.properties.metadata: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["error"]) -> MetaOapg.properties.error: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["id", "stepRuns", "collectionMethod", "slug", "previousRunId", "metadata", "error", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["stepRuns"]) -> MetaOapg.properties.stepRuns: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["collectionMethod"]) -> typing.Union[MetaOapg.properties.collectionMethod, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["slug"]) -> typing.Union[MetaOapg.properties.slug, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["previousRunId"]) -> typing.Union[MetaOapg.properties.previousRunId, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["metadata"]) -> typing.Union[MetaOapg.properties.metadata, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["error"]) -> typing.Union[MetaOapg.properties.error, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["id", "stepRuns", "collectionMethod", "slug", "previousRunId", "metadata", "error", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        stepRuns: typing.Union[MetaOapg.properties.stepRuns, list, tuple, ],
        id: typing.Union[MetaOapg.properties.id, str, uuid.UUID, ],
        collectionMethod: typing.Union[MetaOapg.properties.collectionMethod, str, schemas.Unset] = schemas.unset,
        slug: typing.Union[MetaOapg.properties.slug, str, schemas.Unset] = schemas.unset,
        previousRunId: typing.Union[MetaOapg.properties.previousRunId, None, str, uuid.UUID, schemas.Unset] = schemas.unset,
        metadata: typing.Union[MetaOapg.properties.metadata, dict, frozendict.frozendict, None, schemas.Unset] = schemas.unset,
        error: typing.Union[MetaOapg.properties.error, None, str, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'RunRequest':
        return super().__new__(
            cls,
            *_args,
            stepRuns=stepRuns,
            id=id,
            collectionMethod=collectionMethod,
            slug=slug,
            previousRunId=previousRunId,
            metadata=metadata,
            error=error,
            _configuration=_configuration,
            **kwargs,
        )

from gentrace.model.metadata_value_object import MetadataValueObject
from gentrace.model.step_run import StepRun



# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from gentrace.paths.v2_folders_id import Api

from gentrace.paths import PathValues

path = PathValues.V2_FOLDERS_ID


# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.27.0
    Generated by: https://openapi-generator.tech
"""

from gentrace.paths.v1_feedback.post import V1FeedbackPost
from gentrace.paths.v1_files_upload.post import V1FilesUploadPost
from gentrace.paths.v1_pipelines.get import V1PipelinesGet
from gentrace.paths.v1_run.post import V1RunPost
from gentrace.paths.v1_test_case.get import V1TestCaseGet
from gentrace.paths.v1_test_case.patch import V1TestCasePatch
from gentrace.paths.v1_test_case.post import V1TestCasePost
from gentrace.paths.v1_test_result.get import V1TestResultGet
from gentrace.paths.v1_test_result_id.get import V1TestResultIdGet
from gentrace.paths.v1_test_result_id.post import V1TestResultIdPost
from gentrace.paths.v1_test_result.post import V1TestResultPost
from gentrace.paths.v1_test_result_simple_id.post import V1TestResultSimpleIdPost
from gentrace.paths.v1_test_result_simple.post import V1TestResultSimplePost
from gentrace.paths.v1_test_result_status.get import V1TestResultStatusGet


class V1Api(
    V1FeedbackPost,
    V1FilesUploadPost,
    V1PipelinesGet,
    V1RunPost,
    V1TestCaseGet,
    V1TestCasePatch,
    V1TestCasePost,
    V1TestResultGet,
    V1TestResultIdGet,
    V1TestResultIdPost,
    V1TestResultPost,
    V1TestResultSimpleIdPost,
    V1TestResultSimplePost,
    V1TestResultStatusGet,
):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    pass



import typing_extensions

from gentrace.apis.tags import TagValues
from gentrace.apis.tags.v1_api import V1Api
from gentrace.apis.tags.v2_api import V2Api
from gentrace.apis.tags.v3_api import V3Api
from gentrace.apis.tags.feedback_api import FeedbackApi

TagToApi = typing_extensions.TypedDict(
    'TagToApi',
    {
        TagValues.V1: V1Api,
        TagValues.V2: V2Api,
        TagValues.V3: V3Api,
        TagValues.FEEDBACK: FeedbackApi,
    }
)

tag_to_api = TagToApi(
    {
        TagValues.V1: V1Api,
        TagValues.V2: V2Api,
        TagValues.V3: V3Api,
        TagValues.FEEDBACK: FeedbackApi,
    }
)



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.27.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from gentrace import schemas  # noqa: F401


class MetadataValueObject(
    schemas.ComposedBase,
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "type",
        }
        
        @staticmethod
        def discriminator():
            return {
                'type': {
                    'BooleanType': BooleanType,
                    'StringType': StringType,
                    'UrlType': UrlType,
                    'boolean': BooleanType,
                    'string': StringType,
                    'url': UrlType,
                }
            }
        
        class properties:
            type = schemas.StrSchema
            __annotations__ = {
                "type": type,
            }
        
        @classmethod
        @functools.lru_cache()
        def one_of(cls):
            # we need this here to make our import statements work
            # we must store _composed_schemas in here so the code is only run
            # when we invoke this method. If we kept this at the class
            # level we would get an error because the class level
            # code would be run when this module is imported, and these composed
            # classes don't exist yet because their module has not finished
            # loading
            return [
                UrlType,
                StringType,
                BooleanType,
            ]

    
    type: MetaOapg.properties.type
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["type"]) -> MetaOapg.properties.type: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["type", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["type"]) -> MetaOapg.properties.type: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["type", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        type: typing.Union[MetaOapg.properties.type, str, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'MetadataValueObject':
        return super().__new__(
            cls,
            *_args,
            type=type,
            _configuration=_configuration,
            **kwargs,
        )

from gentrace.model.boolean_type import BooleanType
from gentrace.model.string_type import StringType
from gentrace.model.url_type import UrlType



from typing import Optional, Type, cast

from openai import OpenAI

from gentrace.providers.init import GENTRACE_CONFIG_STATE


class LazyLoader:
    _module: Optional[Type] = None

    @classmethod
    def _load_module(cls) -> Type:
        if cls._module is None:
            from gentrace.providers.llms.openai_v1 import (
                GentraceAsyncOpenAI as _GentraceAsyncOpenAI,
            )
            from gentrace.providers.llms.openai_v1 import (
                GentraceSyncOpenAI as _GentraceSyncOpenAI,
            )
            cls._module = {
                "GentraceSyncOpenAI": _GentraceSyncOpenAI,
                "GentraceAsyncOpenAI": _GentraceAsyncOpenAI,
            }
        return cls._module


class SimpleGentraceSyncOpenAI:
    def __init__(self, *args, **kwargs):
        if not GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"]:
            raise ValueError(
                "No Gentrace API key available. Please use init() to set the API key."
            )

        gentrace_config = {
            "api_key": GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"],
            "host": GENTRACE_CONFIG_STATE["GENTRACE_BASE_PATH"],
        }


        GentraceSyncOpenAI = LazyLoader._load_module()["GentraceSyncOpenAI"]
        self._instance = GentraceSyncOpenAI(*args, **kwargs, gentrace_config=gentrace_config)

    def __getattr__(self, name):
        return getattr(self._instance, name)

SimpleGentraceSyncOpenAITyped = cast(OpenAI, SimpleGentraceSyncOpenAI)


class SimpleGentraceAsyncOpenAI:

    def __init__(self, *args, **kwargs):
        if not GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"]:
            raise ValueError(
                "No Gentrace API key available. Please use init() to set the API key."
            )

        gentrace_config = {
            "api_key": GENTRACE_CONFIG_STATE["GENTRACE_API_KEY"],
            "host": GENTRACE_CONFIG_STATE["GENTRACE_BASE_PATH"],
        }

        GentraceAsyncOpenAI = LazyLoader._load_module()["GentraceAsyncOpenAI"]
        self._instance = GentraceAsyncOpenAI(*args, **kwargs, gentrace_config=gentrace_config)

    def __getattr__(self, name):
        return getattr(self._instance, name)


SimpleGentraceAsyncOpenAITyped = cast(OpenAI, SimpleGentraceAsyncOpenAI)

__all__ = [
    "SimpleGentraceSyncOpenAITyped",
    "SimpleGentraceAsyncOpenAITyped"
]



import copy
import json
import os
import uuid
from datetime import datetime
from itertools import zip_longest
from typing import Any, Callable, Dict, List, Optional, Tuple, TypedDict

from gentrace.api_client import ApiClient
from gentrace.apis.tags.v1_api import V1Api
from gentrace.apis.tags.v2_api import V2Api
from gentrace.model.evaluator_v2 import EvaluatorV2
from gentrace.model.expanded_test_result import ExpandedTestResult
from gentrace.model.pipeline import Pipeline
from gentrace.model.test_case import TestCase
from gentrace.model.test_case_v2 import TestCaseV2
from gentrace.models import TestResult
from gentrace.providers.context import ResultContext
from gentrace.providers.init import (
    GENTRACE_CONFIG_STATE,
)
from gentrace.providers.pipeline_run import PipelineRun, flush
from gentrace.providers.utils import (
    decrement_test_counter,
    get_test_counter,
    increment_test_counter,
)


class Result(TypedDict):
    resultId: str


class TestCaseDict(TypedDict):
    id: str
    createdAt: str
    updatedAt: str
    archivedAt: Optional[str]
    expectedOutputs: Optional[Dict[str, Any]]
    inputs: Dict[str, Any]
    name: str
    pipelineId: str


def is_valid_uuid(val: str):
    try:
        uuid.UUID(str(val))
        return True
    except ValueError:
        return False


def get_evaluators(
        pipeline_id: Optional[str] = None,
        pipeline_slug: Optional[str] = None,
) -> List[EvaluatorV2]:
    """
    Retrieves evaluators  for a given pipeline ID from the Gentrace API.

    Args:
        pipeline_slug (str): The pipeline slug to retrieve evaluators for.
        pipeline_id (str): The ID of the pipeline to retrieve evaluators for.

    Raises:
        ValueError: If the SDK is not initialized. Call init() first.

    Returns:
        list: A list of evaluators.
    """

    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V2Api(api_client=api_client)

    if not pipeline_id and not pipeline_slug:
        pipeline_slug = 'null'  # get template evaluators

    response = api.v2_evaluators_get({
        "pipelineId": pipeline_id,
        "pipelineSlug": pipeline_slug
    })

    evaluators = response.body.get("data", [])

    return evaluators


def get_test_cases(
        pipeline_id: Optional[str] = None,
        pipeline_slug: Optional[str] = None,
        dataset_id: Optional[str] = None,
) -> List[TestCase]:
    """
    Retrieves test cases for a given dataset ID, pipeline ID, or pipeline slug from the Gentrace API.

    Args:
        dataset_id (str): The ID of the dataset to retrieve test cases for.
        pipeline_id (str): The ID of the pipeline to retrieve test cases for.
        pipeline_slug (str): The pipeline slug to retrieve test cases for.

    Raises:
        ValueError: If the SDK is not initialized. Call init() first.
        ValueError: If neither dataset_id, pipeline_id, nor pipeline_slug is provided.

    Returns:
        list: A list of test cases.
    """

    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V1Api(api_client=api_client)

    if not dataset_id and not pipeline_id and not pipeline_slug:
        raise ValueError("Either dataset_id, pipeline_id, or pipeline_slug must be defined.")

    if dataset_id:
        response = api.v1_test_case_get({"datasetId": dataset_id})
    else:
        if pipeline_slug and not is_valid_uuid(pipeline_slug):
            all_pipelines = get_pipelines(slug=pipeline_slug)

            matching_pipeline = next(
                (
                    pipeline
                    for pipeline in all_pipelines
                    if pipeline["slug"] == pipeline_slug
                ),
                None,
            )

            if not matching_pipeline:
                raise ValueError(f"Could not find the specified pipeline ({pipeline_slug})")

            pipeline_id = matching_pipeline.get("id")

        params = {}
        if pipeline_id:
            params["pipelineId"] = pipeline_id
        elif pipeline_slug:
            params["pipelineSlug"] = pipeline_slug

        response = api.v1_test_case_get(params)

    test_cases = response.body.get("testCases", [])
    return test_cases

def get_test_case(
        case_id: str,
) -> TestCaseV2:
    """
    Retrieves a test case for a given test case ID from the Gentrace API.

    Args:
        case_id (str): The test case ID to retrieve.

    Raises:
        ValueError: If the SDK is not initialized. Call init() first.

    Returns:
        TestCaseV2: The test case.
    """

    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V2Api(api_client=api_client)

    response = api.v2_test_cases_id_get({"id": case_id})
    test_case = response.body
    return test_case


class CreateTestCasePayload(TypedDict):
    name: str
    inputs: Dict[str, Any]
    expectedOutputs: Optional[Dict[str, Any]]


class SingleTestCasePayload(TypedDict):
    name: str
    inputs: Dict[str, Any]
    expectedOutputs: Optional[Dict[str, Any]]


class UpdateTestCasePayload(TypedDict):
    id: str
    name: Optional[str]
    inputs: Optional[Dict[str, Any]]
    expectedOutputs: Optional[Dict[str, Any]]
    archived: Optional[bool]


class UpdateTestCaseResponse(TypedDict):
    caseId: str


def create_test_cases(
        pipeline_slug: Optional[str] = None,
        payload: List[TestCaseDict] = [],
        dataset_id: Optional[str] = None
) -> int:
    """Creates multiple test cases for a specified pipeline using the Gentrace API.

    Parameters:
    - pipeline_slug (Optional[str]): The unique identifier of the pipeline to which the test cases should be added.
    - payload (List[TestCaseDict]): The array payload containing the test cases to be created.
    - dataset_id (Optional[str]): The unique identifier of the dataset to which the test cases should be added.

    Returns:
    - int: Count of test cases created.

    Raises:
    - ValueError: If the Gentrace API key is not initialized, if neither pipeline_slug nor dataset_id is provided,
                  or if the payload list is empty.

    Note:
    Ensure that the Gentrace API is initialized by calling init() before using this function.
    If both pipeline_slug and dataset_id are provided, dataset_id will be prioritized.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    if not payload:
        raise ValueError("Payload list cannot be empty")

    api_client = ApiClient(configuration=config)
    api = V1Api(api_client=api_client)

    if not dataset_id and not pipeline_slug:
        raise ValueError("Either pipeline_slug or dataset_id must be provided")

    request_payload = {"testCases": payload}
    if dataset_id:
        request_payload["datasetId"] = dataset_id
    elif pipeline_slug:
        request_payload["pipelineSlug"] = pipeline_slug

    response = api.v1_test_case_post(request_payload)
    count = response.body.get("creationCount", None)
    return count

def create_test_case(
        pipeline_slug: Optional[str] = None,
        payload: Optional[SingleTestCasePayload] = None,
        dataset_id: Optional[str] = None
) -> str:
    """
    Creates a single test case for a specified pipeline or dataset using the Gentrace API.

    Parameters:
    - payload (SingleTestCasePayload): The payload containing the test case to be created.
    - pipeline_slug (Optional[str]): The unique identifier of the pipeline to which the test case should be added.
    - dataset_id (Optional[str]): The unique identifier of the dataset to which the test case should be added.

    Returns:
    - str: The identifier (caseId) of the created test case.

    Raises:
    - ValueError: If the Gentrace API key is not initialized, if neither pipeline_slug nor dataset_id is provided,
                  or if the payload is not provided.

    Note:
    Ensure that the Gentrace API is initialized by calling init() before using this function.
    If both pipeline_slug and dataset_id are provided, dataset_id will be prioritized.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    if payload is None:
        raise ValueError("Payload must be provided")

    api_client = ApiClient(configuration=config)
    api = V1Api(api_client=api_client)

    if not dataset_id and not pipeline_slug:
        raise ValueError("Either pipeline_slug or dataset_id must be provided")

    request_payload = {**payload}
    if dataset_id:
        request_payload["datasetId"] = dataset_id
    elif pipeline_slug:
        request_payload["pipelineSlug"] = pipeline_slug

    response = api.v1_test_case_post(request_payload)
    case_id = response.body.get("caseId", None)
    return case_id


def update_test_case(pipeline_slug: str, payload: UpdateTestCasePayload) -> str:
    """
    Updates a test case for a specified pipeline using the Gentrace API.

    Parameters:
    - pipeline_slug (str): The unique identifier of the pipeline where the test case exists.
    - payload (UpdateTestCasePayload): The payload containing details of the test case to be updated.

    Returns:
    - str: The identifier (caseId) of the updated test case.

    Raises:
    - ValueError: If the Gentrace API key is not initialized or if the pipeline_slug is not passed.

    Note:
    Ensure that the Gentrace API is initialized by calling init() before using this function.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V1Api(api_client=api_client)

    if not pipeline_slug:
        raise ValueError("pipeline_slug must be passed")

    response = api.v1_test_case_patch({"pipelineSlug": pipeline_slug, **payload})
    case_id = response.body.get("caseId", None)
    return case_id


def delete_test_case(test_case_id: str) -> bool:
    """
    Deletes a test case using the Gentrace API.

    Args:
        test_case_id (str): The ID of the test case to delete.

    Raises:
        ValueError: If the Gentrace API key is not initialized.

    Returns:
        bool: True if the test case was successfully deleted, False otherwise.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V2Api(api_client=api_client)

    response = api.v2_test_cases_id_delete(path_params={"id": test_case_id})
    return response.body.get("success", False)


def submit_prepared_test_runs(
        pipeline_slug: str,
        test_runs: List[Dict],
        context: Optional[ResultContext] = None,
        result_name: Optional[str] = None,
) -> Result:
    """
    INTERNAL TO PACKAGE:

    Submits prepared test runs to the Gentrace API for a given pipeline ID. This method requires that you
    create TestRun objects yourself. We recommend using the submit_test_result method instead.

    Args:
        pipeline_slug (str): The pipeline slug
        test_runs (List[Dict]): A list of test runs to submit.
        context (Optional[ResultContext]): Context key pairs
        result_name (str, optional): The name of the test result. Defaults to None.

    Raises:
        ValueError: If the SDK is not initialized. Call init() first.

    Returns:
        Run: The response data from the API.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V1Api(api_client=api_client)

    for test_run in test_runs:
        test_run["inputs"] = (
            json.dumps(test_run["inputs"])
            if isinstance(test_run["inputs"], str)
            else test_run["inputs"]
        )

    params = construct_submission_payload(
        pipeline_slug, test_runs, context, result_name
    )
    response = api.v1_test_result_simple_post(params)
    return response.body


def construct_submission_payload(
        pipeline_identifier: str,
        test_runs: List[Dict],
        context: Optional[ResultContext] = None,
        result_name: Optional[str] = None,
):
    """
    Constructs a dictionary payload for submitting test runs to a server.

    Args:
        pipeline_identifier (str): The pipeline slug, or pipeline ID
        test_results (List[Dict]): A list of dictionaries containing test results.
        context (Optional[ResultContext]): Context key pairs
        result_name (str, optional): The name of the test result. Defaults to None.

    Returns:
        Dict: A dictionary payload containing the pipeline slug, test runs, and optional branch and commit information.
    """
    params = {
        "testRuns": test_runs,
    }

    if is_valid_uuid(pipeline_identifier):
        params["pipelineId"] = pipeline_identifier
    else:
        params["pipelineSlug"] = pipeline_identifier

    if GENTRACE_CONFIG_STATE["GENTRACE_RUN_NAME"]:
        params["name"] = GENTRACE_CONFIG_STATE["GENTRACE_RUN_NAME"]

    if GENTRACE_CONFIG_STATE["GENTRACE_RESULT_NAME"]:
        params["name"] = GENTRACE_CONFIG_STATE["GENTRACE_RESULT_NAME"]

    if result_name:
        params["name"] = result_name

    if os.getenv("GENTRACE_BRANCH") or GENTRACE_CONFIG_STATE["GENTRACE_BRANCH"]:
        params["branch"] = GENTRACE_CONFIG_STATE["GENTRACE_BRANCH"] or os.getenv(
            "GENTRACE_BRANCH"
        )

    if os.getenv("GENTRACE_COMMIT") or GENTRACE_CONFIG_STATE["GENTRACE_COMMIT"]:
        params["commit"] = GENTRACE_CONFIG_STATE["GENTRACE_COMMIT"] or os.getenv(
            "GENTRACE_COMMIT"
        )

    if context and context.get("metadata"):
        params["metadata"] = context.get("metadata")

    return params


class OutputStep(TypedDict):
    key: str
    output: str
    inputs: Optional[Dict[str, Any]]


def submit_test_result(
        pipeline_slug: str,
        test_cases: List[TestCase],
        outputs_list: List[Dict[str, Any]],
        context: Optional[ResultContext] = None,
        result_name: Optional[str] = None,
) -> Result:
    """
    Submits a test result by creating TestRun objects from given test cases and corresponding outputs.
    To use a Gentrace runner to capture intermediate steps, use run_test instead.

    Args:
        pipeline_slug (str): The pipeline slug
        test_cases (List[TestCase]): A list of TestCase objects.
        outputs_list (List[Dict[str, Any]]): A list of outputs corresponding to each TestCase.
        result_name (str, optional): The name of the test result. Defaults to None.

    Raises:
        ValueError: If the Gentrace API key is not initialized.

    Returns:
        Run: The response data from the test_result_simple_post SDK method.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    if len(test_cases) != len(outputs_list):
        raise ValueError("`test_cases` and `outputs` should be the same length.")

    if is_valid_uuid(pipeline_slug):
        raise ValueError("You passed a pipeline ID instead of a pipeline slug.")

    test_runs = []

    for test_case, outputs in zip_longest(test_cases, outputs_list, fillvalue=None):
        result = {
            "caseId": test_case["id"],
            "inputs": (
                json.loads(test_case["inputs"])
                if isinstance(test_case["inputs"], str)
                else test_case["inputs"]
            ),
            "outputs": outputs,
        }

        test_runs.append(result)

    return submit_prepared_test_runs(pipeline_slug, test_runs, context, result_name)


def get_pipelines(
        label: Optional[str] = None,
        slug: Optional[str] = None,
) -> List[Pipeline]:
    """
    Get pipelines from the Gentrace API, optionally filtered by label or by slug

    Args:
        label (str, optional): The label of the test set. Defaults to None.
        slug (str, optional): The slug of the test set. Defaults to None.

    Raises:
        ValueError: If the Gentrace API key is not initialized.

    Returns:
        Run: The array of test sets returned by the Gentrace API.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V1Api(api_client=api_client)

    params = {}

    if label:
        params["label"] = label

    if slug:
        params["slug"] = slug

    response = api.v1_pipelines_get(params)

    pipelines = response.body.get("pipelines")

    return pipelines


def get_test_result(result_id: str) -> ExpandedTestResult:
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V1Api(api_client=api_client)

    params = {}

    if result_id:
        params["id"] = result_id

    response = api.v1_test_result_id_get(params)

    return response.body


def get_test_results(
        pipeline_slug: str,
) -> List[TestResult]:
    """
    Fetches test results using the Gentrace API.

    Args:
        pipeline_slug (str): The pipeline slug to pull test results

    Returns:
        List[TestResult]: A list of test results fetched from the Gentrace API.

    Raises:
        ValueError: If the Gentrace API key is not initialized. Call `init()` function first.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V1Api(api_client=api_client)

    params = {}

    if pipeline_slug:
        params["pipelineSlug"] = pipeline_slug

    response = api.v1_test_result_get(params)

    test_results = response.body.get("testResults")

    return test_results


class EvaluationDictBase(TypedDict, total=True):
    evaluatorId: str
    runId: str
    note: Optional[str]


class EvaluationDict(EvaluationDictBase, total=False):
    evalValue: Optional[float]
    evalLabel: Optional[str]


def bulk_create_evaluations(
        payloads: List[EvaluationDict],
):
    """
    Creates multiple evaluations using the Gentrace API.

    Args:
        payloads (List[EvaluationDict]): The array payload containing the evaluations to be created.

    Returns:
        int: Count of evaluations created.

    Raises:
        ValueError: If the Gentrace API key is not initialized.

    Note:
    Ensure that the Gentrace API is initialized by calling init() before using this function.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V2Api(api_client=api_client)

    result = api.v2_evaluations_bulk_post({"data": payloads})

    count = result.body.get("count", None)
    return count

def run_test(
        pipeline_slug: str,
        handler,
        context: Optional[ResultContext] = None,
        case_filter: Optional[Callable[[TestCase], bool]] = None,
        result_name: Optional[str] = None,
        dataset_id: Optional[str] = None,
) -> Result:
    """
    Runs a test by pulling down test cases from Gentrace, running them through †he
    provided callback (once per test case), and submitting the result report back to Gentrace.

    Args:
        pipeline_slug (str): The slug of the pipeline to run.
        handler (Callable[[TestCase], List[Dict]]): A function that takes a TestCase and returns a tuple of
          the output and a PipelineRun class instance that contains the list of steps taken by the pipeline.
        context (Optional[ResultContext]): Context key pairs
        case_filter: Optional[Callable[[TestCase], bool]] = None
        result_name (str, optional): The name of the test result. Defaults to None.
        dataset_id (str, optional): The ID of the dataset to retrieve test cases for. Defaults to None.

    Raises:
        ValueError: If the Gentrace API key is not initialized.

    Returns:
        Response data from the Gentrace API's /test-result POST method. This should just be a dictionary
        similar to the following:

        {
            "resultId": "161c623d-ee92-417f-823a-cf9f7eccf557",
        }
    """
    increment_test_counter()

    try:
        config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
        if not config:
            raise ValueError("Gentrace API key not initialized. Call init() first.")

        api_client = ApiClient(configuration=config)
        api = V1Api(api_client=api_client)

        all_pipelines = get_pipelines()

        matching_pipeline = next(
            (
                pipeline
                for pipeline in all_pipelines
                if pipeline["slug"] == pipeline_slug
            ),
            None,
        )

        if not matching_pipeline:
            raise ValueError(f"Could not find the specified pipeline ({pipeline_slug})")

        if dataset_id:
            test_cases = get_test_cases(dataset_id=dataset_id)
        else:
            test_cases = get_test_cases(pipeline_id=matching_pipeline["id"])

        test_runs = []

        for test_case in test_cases:
            if case_filter and not case_filter(test_case):
                continue

            [output, pipeline_run] = handler(test_case)

            merged_metadata = {}

            step_runs_data = []
            for step_run in pipeline_run.step_runs:
                # Extract metadata without mutating original contexts
                this_context = copy.deepcopy(pipeline_run.context)
                this_context_metadata = this_context.get("metadata", {})
                step_run_context = copy.deepcopy(step_run.context)
                step_run_context_metadata = step_run_context.get("metadata", {})

                merged_metadata.update(this_context_metadata)
                merged_metadata.update(step_run_context_metadata)

                this_context.pop("metadata", None)
                step_run_context.pop("metadata", None)

                this_context.pop("previousRunId", None)
                step_run_context.pop("previousRunId", None)

                step_runs_data.append(
                    {
                        "providerName": step_run.provider,
                        "invocation": step_run.invocation,
                        "modelParams": step_run.model_params,
                        "inputs": step_run.inputs,
                        "outputs": step_run.outputs,
                        "elapsedTime": step_run.elapsed_time,
                        "startTime": step_run.start_time,
                        "endTime": step_run.end_time,
                        "context": {**this_context, **step_run_context},
                    }
                )

            test_run = {
                "caseId": test_case["id"],
                "metadata": merged_metadata,
                "previousRunId": pipeline_run.context.get("previousRunId"),
                "stepRuns": step_runs_data,
            }

            if pipeline_run.get_id():
                test_run["id"] = pipeline_run.get_id()

            test_runs.append(test_run)

        params = construct_submission_payload(
            matching_pipeline["id"], test_runs, context, result_name
        )
        params["collectionMethod"] = "runner"

        response = api.v1_test_result_post(params)

        return response.body
    except Exception as e:
        raise e
    finally:
        decrement_test_counter()

        # OK to flush here and introduce more latency since this is just used for test anyway
        flush()


def get_test_runners(
    pipeline: Pipeline,
    dataset_id: Optional[str] = None,
    case_filter: Optional[Callable[[Dict[str, Any]], bool]] = None
) -> List[Tuple[PipelineRun, Dict[str, Any]]]:
    """
    Retrieves test runners for a given pipeline

    Args:
        pipeline (Pipeline): The pipeline instance
        dataset_id (Optional[str]): Optional dataset ID to filter test cases by.
        case_filter (Optional[Callable[[Dict[str, Any]], bool]]): Optional function to filter test cases

    Raises:
        ValueError: If the Gentrace API key is not initialized.

    Returns:
        A list of (PipelineRun, TestCase) tuples
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V1Api(api_client=api_client)

    if not pipeline:
        raise ValueError("Invalid pipeline found")

    params = {}
    if dataset_id:
        params["datasetId"] = dataset_id
    if is_valid_uuid(pipeline.id):
        params["pipelineId"] = pipeline.id
    else:
        params["pipelineSlug"] = pipeline.slug

    response = api.v1_test_case_get(params)
    test_cases = response.body.get("testCases", [])

    test_runners = []
    total_test_cases = len(test_cases)
    filtered_test_cases = 0

    for test_case in test_cases:
        if case_filter and not case_filter(test_case):
            continue

        filtered_test_cases += 1
        pipeline_run = pipeline.start()
        test_runners.append((pipeline_run, test_case))

    print(f"Total test cases: {total_test_cases}")
    print(f"Filtered test cases: {filtered_test_cases}")
    return test_runners

def submit_test_runners(
        pipeline: Pipeline,
        pipeline_run_test_cases: List[Tuple[PipelineRun, TestCase]],
        context: Optional[ResultContext] = None,
        case_filter: Optional[Callable[[TestCase], bool]] = None,
        result_name: Optional[str] = None,
) -> Result:
    """
    Submits test runners for a given pipeline

    Args:
        pipeline (Pipeline): The pipeline instance
        pipeline_run_test_cases (List[Tuple[PipelineRun, TestCase]]): A list of (PipelineRun, TestCase) tuples
        context (Optional[ResultContext]): Context key pairs
        case_filter: Optional[Callable[[TestCase], bool]] = None
        result_name (str, optional): The name of the test result. Defaults to None.

    Raises:
        ValueError: If the Gentrace API key is not initialized.

    Returns:
        Response data from the Gentrace API's /test-result POST method. This should just be a dictionary
        similar to the following:

        {
            "resultId": "161c623d-ee92-417f-823a-cf9f7eccf557",
        }
    """
    try:
        config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
        if not config:
            raise ValueError("Gentrace API key not initialized. Call init() first.")

        api_client = ApiClient(configuration=config)
        api = V1Api(api_client=api_client)

        if not pipeline:
            raise ValueError("Invalid pipeline found")

        test_runs = []

        for pipeline_run, test_case in pipeline_run_test_cases:
            if case_filter and not case_filter(test_case):
                continue

            merged_metadata = {}

            step_runs_data = []
            for step_run in pipeline_run.step_runs:
                # Extract metadata without mutating original contexts
                this_context = copy.deepcopy(pipeline_run.context)
                this_context_metadata = this_context.get("metadata", {})
                step_run_context = copy.deepcopy(step_run.context)
                step_run_context_metadata = step_run_context.get("metadata", {})

                merged_metadata.update(this_context_metadata)
                merged_metadata.update(step_run_context_metadata)

                this_context.pop("metadata", None)
                step_run_context.pop("metadata", None)

                this_context.pop("previousRunId", None)
                step_run_context.pop("previousRunId", None)

                step_runs_data.append(
                    {
                        "providerName": step_run.provider,
                        "invocation": step_run.invocation,
                        "modelParams": step_run.model_params,
                        "inputs": step_run.inputs,
                        "outputs": step_run.outputs,
                        "elapsedTime": step_run.elapsed_time,
                        "startTime": step_run.start_time,
                        "endTime": step_run.end_time,
                        "context": {**this_context, **step_run_context},
                    }
                )

            test_run = {
                "caseId": test_case["id"],
                "metadata": merged_metadata,
                "previousRunId": pipeline_run.context.get("previousRunId"),
                "stepRuns": step_runs_data,
            }

            if pipeline_run.get_id():
                test_run["id"] = pipeline_run.get_id()

            test_runs.append(test_run)

        params = construct_submission_payload(
            pipeline.id or pipeline.slug, test_runs, context, result_name
        )
        params["collectionMethod"] = "runner"

        response = api.v1_test_result_post(params)
        return response.body
    except Exception as e:
        raise e


class DatasetV2(TypedDict):
    id: str
    name: str
    description: Optional[str]
    pipelineId: str
    createdAt: datetime
    updatedAt: datetime
    archivedAt: Optional[datetime]


class CreateDatasetV2(TypedDict):
    name: str
    description: Optional[str]
    pipelineId: str


class UpdateDatasetV2(TypedDict, total=False):
    name: Optional[str]
    description: Optional[str]
    archived: Optional[bool]


class DatasetListResponse(TypedDict):
    data: List[DatasetV2]


def get_datasets(
        pipeline_slug: Optional[str] = None,
        pipeline_id: Optional[str] = None,
        archived: Optional[bool] = None
) -> DatasetListResponse:
    """
    Get datasets from the Gentrace API, optionally filtered by pipeline slug, pipeline ID, or archived status.

    Args:
        pipeline_slug (str, optional): The slug of the pipeline to filter datasets by. Defaults to None.
        pipeline_id (str, optional): The ID of the pipeline to filter datasets by. Defaults to None.
        archived (bool, optional): Filter datasets by archived status. Defaults to None.

    Raises:
        ValueError: If the Gentrace API key is not initialized.

    Returns:
        DatasetListResponse: The response containing an array of datasets returned by the Gentrace API.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V2Api(api_client=api_client)

    query_params = {}
    if pipeline_slug:
        query_params["pipelineSlug"] = pipeline_slug
    if pipeline_id:
        query_params["pipelineId"] = pipeline_id
    if archived is not None:
        query_params["archived"] = archived

    response = api.v2_datasets_get(
        query_params=query_params,
    )
    return DatasetListResponse(data=response.body.get("data", []))


def create_dataset(dataset_data: CreateDatasetV2) -> DatasetV2:
    """
    Create a new dataset using the Gentrace API.

    Args:
        dataset_data (CreateDatasetV2): The data for creating the new dataset.

    Raises:
        ValueError: If the Gentrace API key is not initialized.

    Returns:
        DatasetV2: The created dataset returned by the Gentrace API.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V2Api(api_client=api_client)

    response = api.v2_datasets_post(
        body=dataset_data,
    )
    return DatasetV2(**response.body)


def get_dataset(dataset_id: str) -> DatasetV2:
    """
    Get a single dataset from the Gentrace API by its ID.

    Args:
        dataset_id (str): The ID of the dataset to retrieve.

    Raises:
        ValueError: If the Gentrace API key is not initialized.

    Returns:
        DatasetV2: The dataset returned by the Gentrace API.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V2Api(api_client=api_client)

    response = api.v2_datasets_id_get(
        path_params={"id": dataset_id}
    )
    return DatasetV2(**response.body)


def update_dataset(dataset_id: str, update_data: UpdateDatasetV2) -> DatasetV2:
    """
    Update an existing dataset using the Gentrace API.

    Args:
        dataset_id (str): The ID of the dataset to update.
        update_data (UpdateDatasetV2): The data for updating the dataset.

    Raises:
        ValueError: If the Gentrace API key is not initialized.

    Returns:
        DatasetV2: The updated dataset returned by the Gentrace API.
    """
    config = GENTRACE_CONFIG_STATE["global_gentrace_config"]
    if not config:
        raise ValueError("Gentrace API key not initialized. Call init() first.")

    api_client = ApiClient(configuration=config)
    api = V2Api(api_client=api_client)

    response = api.v2_datasets_id_post(
        body=update_data,
        path_params={"id": dataset_id}
    )
    return DatasetV2(**response.body)


__all__ = [
    "get_evaluators",
    "get_test_cases",
    "get_test_case",
    "create_test_cases",
    "create_test_case",
    "delete_test_case",
    "get_test_results",
    "get_test_result",
    "update_test_case",
    "submit_test_result",
    "get_pipelines",
    "construct_submission_payload",
    "run_test",
    "get_test_runners",
    "submit_test_runners",
    "bulk_create_evaluations",
    "OutputStep",
    "EvaluationDict",
    "get_dataset",
    "update_dataset",
    "create_dataset",
    "get_datasets",
]



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.27.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from gentrace import schemas  # noqa: F401


class TestResult(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "createdAt",
            "id",
            "pipelineId",
            "updatedAt",
        }
        
        class properties:
            id = schemas.UUIDSchema
            createdAt = schemas.DateTimeSchema
            updatedAt = schemas.DateTimeSchema
            pipelineId = schemas.UUIDSchema
            
            
            class branch(
                schemas.StrBase,
                schemas.NoneBase,
                schemas.Schema,
                schemas.NoneStrMixin
            ):
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[None, str, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'branch':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                    )
            
            
            class commit(
                schemas.StrBase,
                schemas.NoneBase,
                schemas.Schema,
                schemas.NoneStrMixin
            ):
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[None, str, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'commit':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                    )
            
            
            class metadata(
                schemas.DictBase,
                schemas.NoneBase,
                schemas.Schema,
                schemas.NoneFrozenDictMixin
            ):
            
            
                class MetaOapg:
                    
                    @staticmethod
                    def additional_properties() -> typing.Type['MetadataValueObject']:
                        return MetadataValueObject
            
                
                def __getitem__(self, name: typing.Union[str, ]) -> 'MetadataValueObject':
                    # dict_instance[name] accessor
                    return super().__getitem__(name)
                
                def get_item_oapg(self, name: typing.Union[str, ]) -> 'MetadataValueObject':
                    return super().get_item_oapg(name)
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, None, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: 'MetadataValueObject',
                ) -> 'metadata':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class name(
                schemas.StrBase,
                schemas.NoneBase,
                schemas.Schema,
                schemas.NoneStrMixin
            ):
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[None, str, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'name':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                    )
            __annotations__ = {
                "id": id,
                "createdAt": createdAt,
                "updatedAt": updatedAt,
                "pipelineId": pipelineId,
                "branch": branch,
                "commit": commit,
                "metadata": metadata,
                "name": name,
            }
    
    createdAt: MetaOapg.properties.createdAt
    id: MetaOapg.properties.id
    pipelineId: MetaOapg.properties.pipelineId
    updatedAt: MetaOapg.properties.updatedAt
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["createdAt"]) -> MetaOapg.properties.createdAt: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["updatedAt"]) -> MetaOapg.properties.updatedAt: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["pipelineId"]) -> MetaOapg.properties.pipelineId: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["branch"]) -> MetaOapg.properties.branch: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["commit"]) -> MetaOapg.properties.commit: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["metadata"]) -> MetaOapg.properties.metadata: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["name"]) -> MetaOapg.properties.name: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["id", "createdAt", "updatedAt", "pipelineId", "branch", "commit", "metadata", "name", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["createdAt"]) -> MetaOapg.properties.createdAt: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["updatedAt"]) -> MetaOapg.properties.updatedAt: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["pipelineId"]) -> MetaOapg.properties.pipelineId: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["branch"]) -> typing.Union[MetaOapg.properties.branch, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["commit"]) -> typing.Union[MetaOapg.properties.commit, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["metadata"]) -> typing.Union[MetaOapg.properties.metadata, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["name"]) -> typing.Union[MetaOapg.properties.name, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["id", "createdAt", "updatedAt", "pipelineId", "branch", "commit", "metadata", "name", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        createdAt: typing.Union[MetaOapg.properties.createdAt, str, datetime, ],
        id: typing.Union[MetaOapg.properties.id, str, uuid.UUID, ],
        pipelineId: typing.Union[MetaOapg.properties.pipelineId, str, uuid.UUID, ],
        updatedAt: typing.Union[MetaOapg.properties.updatedAt, str, datetime, ],
        branch: typing.Union[MetaOapg.properties.branch, None, str, schemas.Unset] = schemas.unset,
        commit: typing.Union[MetaOapg.properties.commit, None, str, schemas.Unset] = schemas.unset,
        metadata: typing.Union[MetaOapg.properties.metadata, dict, frozendict.frozendict, None, schemas.Unset] = schemas.unset,
        name: typing.Union[MetaOapg.properties.name, None, str, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'TestResult':
        return super().__new__(
            cls,
            *_args,
            createdAt=createdAt,
            id=id,
            pipelineId=pipelineId,
            updatedAt=updatedAt,
            branch=branch,
            commit=commit,
            metadata=metadata,
            name=name,
            _configuration=_configuration,
            **kwargs,
        )

from gentrace.model.metadata_value_object import MetadataValueObject



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.27.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from gentrace import schemas  # noqa: F401


class UnixSecondsNullable(
    schemas.Float32Base,
    schemas.NumberBase,
    schemas.NoneBase,
    schemas.Schema,
    schemas.NoneDecimalMixin
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Timestamp in seconds since the UNIX epoch. Can be transformed into a Date object.
    """


    class MetaOapg:
        format = 'float'


    def __new__(
        cls,
        *_args: typing.Union[None, decimal.Decimal, int, float, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
    ) -> 'UnixSecondsNullable':
        return super().__new__(
            cls,
            *_args,
            _configuration=_configuration,
        )



# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.27.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from gentrace import schemas  # noqa: F401


class LocalEvaluationDebug(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        
        class properties:
            resolvedPrompt = schemas.StrSchema
            response = schemas.StrSchema
            finalClassification = schemas.StrSchema
            
            
            class processorLogs(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    
                    
                    class items(
                        schemas.ListSchema
                    ):
                    
                    
                        class MetaOapg:
                            items = schemas.AnyTypeSchema
                    
                        def __new__(
                            cls,
                            _arg: typing.Union[typing.Tuple[typing.Union[MetaOapg.items, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ]], typing.List[typing.Union[MetaOapg.items, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ]]],
                            _configuration: typing.Optional[schemas.Configuration] = None,
                        ) -> 'items':
                            return super().__new__(
                                cls,
                                _arg,
                                _configuration=_configuration,
                            )
                    
                        def __getitem__(self, i: int) -> MetaOapg.items:
                            return super().__getitem__(i)
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[MetaOapg.items, list, tuple, ]], typing.List[typing.Union[MetaOapg.items, list, tuple, ]]],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'processorLogs':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)
            
            
            class logs(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    
                    
                    class items(
                        schemas.ListSchema
                    ):
                    
                    
                        class MetaOapg:
                            items = schemas.AnyTypeSchema
                    
                        def __new__(
                            cls,
                            _arg: typing.Union[typing.Tuple[typing.Union[MetaOapg.items, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ]], typing.List[typing.Union[MetaOapg.items, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ]]],
                            _configuration: typing.Optional[schemas.Configuration] = None,
                        ) -> 'items':
                            return super().__new__(
                                cls,
                                _arg,
                                _configuration=_configuration,
                            )
                    
                        def __getitem__(self, i: int) -> MetaOapg.items:
                            return super().__getitem__(i)
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[MetaOapg.items, list, tuple, ]], typing.List[typing.Union[MetaOapg.items, list, tuple, ]]],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'logs':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)
            
            
            class error(
                schemas.DictSchema
            ):
            
            
                class MetaOapg:
                    
                    class properties:
                        message = schemas.StrSchema
                        date = schemas.DateTimeSchema
                        __annotations__ = {
                            "message": message,
                            "date": date,
                        }
                
                @typing.overload
                def __getitem__(self, name: typing_extensions.Literal["message"]) -> MetaOapg.properties.message: ...
                
                @typing.overload
                def __getitem__(self, name: typing_extensions.Literal["date"]) -> MetaOapg.properties.date: ...
                
                @typing.overload
                def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
                
                def __getitem__(self, name: typing.Union[typing_extensions.Literal["message", "date", ], str]):
                    # dict_instance[name] accessor
                    return super().__getitem__(name)
                
                
                @typing.overload
                def get_item_oapg(self, name: typing_extensions.Literal["message"]) -> typing.Union[MetaOapg.properties.message, schemas.Unset]: ...
                
                @typing.overload
                def get_item_oapg(self, name: typing_extensions.Literal["date"]) -> typing.Union[MetaOapg.properties.date, schemas.Unset]: ...
                
                @typing.overload
                def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
                
                def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["message", "date", ], str]):
                    return super().get_item_oapg(name)
                
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, ],
                    message: typing.Union[MetaOapg.properties.message, str, schemas.Unset] = schemas.unset,
                    date: typing.Union[MetaOapg.properties.date, str, datetime, schemas.Unset] = schemas.unset,
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'error':
                    return super().__new__(
                        cls,
                        *_args,
                        message=message,
                        date=date,
                        _configuration=_configuration,
                        **kwargs,
                    )
            __annotations__ = {
                "resolvedPrompt": resolvedPrompt,
                "response": response,
                "finalClassification": finalClassification,
                "processorLogs": processorLogs,
                "logs": logs,
                "error": error,
            }
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["resolvedPrompt"]) -> MetaOapg.properties.resolvedPrompt: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["response"]) -> MetaOapg.properties.response: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["finalClassification"]) -> MetaOapg.properties.finalClassification: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["processorLogs"]) -> MetaOapg.properties.processorLogs: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["logs"]) -> MetaOapg.properties.logs: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["error"]) -> MetaOapg.properties.error: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["resolvedPrompt", "response", "finalClassification", "processorLogs", "logs", "error", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["resolvedPrompt"]) -> typing.Union[MetaOapg.properties.resolvedPrompt, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["response"]) -> typing.Union[MetaOapg.properties.response, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["finalClassification"]) -> typing.Union[MetaOapg.properties.finalClassification, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["processorLogs"]) -> typing.Union[MetaOapg.properties.processorLogs, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["logs"]) -> typing.Union[MetaOapg.properties.logs, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["error"]) -> typing.Union[MetaOapg.properties.error, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["resolvedPrompt", "response", "finalClassification", "processorLogs", "logs", "error", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        resolvedPrompt: typing.Union[MetaOapg.properties.resolvedPrompt, str, schemas.Unset] = schemas.unset,
        response: typing.Union[MetaOapg.properties.response, str, schemas.Unset] = schemas.unset,
        finalClassification: typing.Union[MetaOapg.properties.finalClassification, str, schemas.Unset] = schemas.unset,
        processorLogs: typing.Union[MetaOapg.properties.processorLogs, list, tuple, schemas.Unset] = schemas.unset,
        logs: typing.Union[MetaOapg.properties.logs, list, tuple, schemas.Unset] = schemas.unset,
        error: typing.Union[MetaOapg.properties.error, dict, frozendict.frozendict, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'LocalEvaluationDebug':
        return super().__new__(
            cls,
            *_args,
            resolvedPrompt=resolvedPrompt,
            response=response,
            finalClassification=finalClassification,
            processorLogs=processorLogs,
            logs=logs,
            error=error,
            _configuration=_configuration,
            **kwargs,
        )



import copy
from typing import Any, Optional

from gentrace.providers.context import Context


class StepRun:
    def __init__(
            self,
            provider: str,
            invocation: str,
            elapsed_time: float,
            start_time: str,
            end_time: str,
            inputs: any,
            model_params: any,
            outputs: any,
            context: Optional[Context] = None,
            error: Optional[str] = None,
    ):
        self.provider = provider
        self.invocation = invocation
        self.elapsed_time = elapsed_time
        self.start_time = start_time
        self.end_time = end_time
        self.inputs = self._convert_to_dict(inputs)
        self.model_params = self._convert_to_dict(model_params)
        self.outputs = self._convert_to_dict(outputs)
        self.context = self._convert_to_dict(context or {})
        self.error = error

    def _convert_to_dict(self, obj: Any) -> Any:
        if hasattr(obj, '__dict__'):
            # Check for DynamicSchema Gentrace custom types
            if hasattr(obj, 'from_openapi_data_oapg'):
                # Primitive type
                if not hasattr(obj, 'items'):
                    return obj
                obj = {k: self._convert_to_dict(v) for k, v in obj.items()}
                return obj

            # Check for model_dump first (works for both v1 and v2)
            if hasattr(obj, 'model_dump'):
                return obj.model_dump()

            # Fallback to dict() if model_dump is not available
            elif hasattr(obj, 'dict'):
                return obj.dict()

            # If neither method is available, use vars() to extrace the object's attributes
            else:
                return vars(obj)
        elif isinstance(obj, dict):
            return {k: self._convert_to_dict(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self._convert_to_dict(item) for item in obj]
        return obj



from gentrace.paths.v1_test_case.get import ApiForget
from gentrace.paths.v1_test_case.post import ApiForpost
from gentrace.paths.v1_test_case.patch import ApiForpatch


class V1TestCase(
    ApiForget,
    ApiForpost,
    ApiForpatch,
):
    pass



import copy
import time
import uuid
from typing import Dict, Optional

import openai
import pystache

from gentrace.configuration import Configuration
from gentrace.providers.pipeline import Pipeline
from gentrace.providers.pipeline_run import PipelineRun
from gentrace.providers.step_run import StepRun
from gentrace.providers.utils import to_date_string


class OpenAIPipelineHandler:
    pipeline_run: Optional[PipelineRun] = None
    pipeline: Optional[Pipeline] = None
    gentrace_config: Dict
    config: Dict = {}

    def __init__(self, gentrace_config, config=None, pipeline=None, pipeline_run=None):
        self.pipeline = pipeline
        self.pipeline_run = pipeline_run
        self.gentrace_config = gentrace_config
        if config:
            self.config = config

    @classmethod
    def setup(cls, config):
        if config:
            for key, value in config.items():
                setattr(openai, key, value)


def create_completion_step_run(
        cls,
        pipeline_slug: str,
        gentrace_config: Configuration,
        start_time,
        end_time,
        base_completion_options,
        prompt_template,
        prompt_inputs,
        completion,
        pipeline_run_id: Optional[str] = None,
        stream=False,
        context={},
        error: Optional[str] = None,
):
    elapsed_time = int((end_time - start_time) * 1000)

    user = base_completion_options.get("user")
    suffix = base_completion_options.get("suffix")
    partial_model_params = {
        k: v for k, v in base_completion_options.items() if k not in ["user", "suffix"]
    }

    inputs_dict = {"prompt": prompt_inputs}
    if user is not None:
        inputs_dict["user"] = user
    if suffix is not None:
        inputs_dict["suffix"] = suffix

    pipeline_run = cls.pipeline_run if hasattr(cls, "pipeline_run") else None

    is_self_contained = not pipeline_run and pipeline_slug

    if is_self_contained:
        pipeline = Pipeline(
            slug=pipeline_slug,
            api_key=gentrace_config.access_token,
            host=gentrace_config.host,
        )

        pipeline_run = PipelineRun(
            pipeline=pipeline, id=pipeline_run_id, context=context
        )

    if pipeline_run:
        pipeline_run.add_step_run(
            OpenAICreateCompletionStepRun(
                elapsed_time,
                to_date_string(start_time),
                to_date_string(end_time),
                inputs_dict,
                {**partial_model_params, "promptTemplate": prompt_template},
                completion,
                context,
                error,
            )
        )

        if is_self_contained:
            submit_result = pipeline_run.submit()

            if not stream:
                completion["pipelineRunId"] = (
                    submit_result["pipelineRunId"]
                    if "pipelineRunId" in submit_result
                    else None
                )


def create_completion_stream_response(stream_list):
    final_response_string = ""
    for value in stream_list:
        if "choices" in value and value["choices"]:
            first_choice = value["choices"][0]
            if "text" in first_choice:
                final_response_string += first_choice["text"]
            elif (
                    "delta" in first_choice
                    and first_choice["delta"]
                    and "content" in first_choice["delta"]
            ):
                final_response_string += first_choice["delta"]["content"]
            elif (
                    "finish_reason" in first_choice
                    and first_choice["finish_reason"] == "stop"
            ):
                break

    final_response = {
        "choices": [
            {
                "finish_reason": None,
                "index": 0,
                "logprobs": None,
                "text": final_response_string,
            }
        ]
    }

    return final_response


def create_chat_completion_stream_response(stream_list):
    final_response_string = ""
    for value in stream_list:
        if "choices" in value and value["choices"]:
            first_choice = value["choices"][0]
            if "text" in first_choice:
                final_response_string += first_choice["text"]
            elif (
                    "delta" in first_choice
                    and first_choice["delta"]
                    and "content" in first_choice["delta"]
            ):
                final_response_string += first_choice["delta"]["content"]
            elif (
                    "finish_reason" in first_choice
                    and first_choice["finish_reason"] == "stop"
            ):
                break

    final_response = {
        "choices": [
            {
                "finish_reason": None,
                "index": 0,
                "message": {"content": final_response_string, "role": "assistant"},
            }
        ]
    }

    return final_response


def intercept_completion(original_fn, gentrace_config: Configuration):
    @classmethod
    def wrapper(cls, *args, **kwargs):
        prompt_template = kwargs.get("prompt_template")
        prompt_inputs = kwargs.get("prompt_inputs")
        prompt = kwargs.get("prompt")
        # @deprecated: pipeline_id is deprecated in favor of pipeline_slug
        pipeline_id = kwargs.pop("pipeline_id", None)
        pipeline_slug = kwargs.pop("pipeline_slug", None)
        context = kwargs.pop("gentrace", {})
        stream = kwargs.get("stream")
        base_completion_options = {
            k: v
            for k, v in kwargs.items()
            if k not in ["prompt_template", "prompt_inputs"]
        }

        effective_pipeline_slug = pipeline_slug or pipeline_id

        if stream:
            rendered_prompt = prompt

            if prompt_template and prompt_inputs:
                rendered_prompt = pystache.render(prompt_template, prompt_inputs)
            else:
                # If no prompt template or inputs, then specify the prompt as the entire input
                prompt_inputs = prompt

            new_completion_options = {
                **base_completion_options,
                "prompt": rendered_prompt,
            }

            start_time = time.time()
            completion = original_fn(**new_completion_options)

            is_self_contained = (
                    not hasattr(cls, "pipeline_run") and effective_pipeline_slug
            )
            if is_self_contained:
                pipeline_run_id = str(uuid.uuid4())

            def profiled_completion():
                modified_response = []
                for value in completion:
                    if value and is_self_contained:
                        value["pipelineRunId"] = pipeline_run_id
                    modified_response.append(value)
                    yield value

                end_time = time.time()

                full_response = create_completion_stream_response(modified_response)

                create_completion_step_run(
                    cls,
                    effective_pipeline_slug,
                    gentrace_config,
                    start_time,
                    end_time,
                    base_completion_options,
                    prompt_template,
                    prompt_inputs,
                    full_response,
                    pipeline_run_id if is_self_contained else None,
                    stream,
                    context,
                )

            return profiled_completion()

        rendered_prompt = prompt

        if prompt_template and prompt_inputs:
            rendered_prompt = pystache.render(prompt_template, prompt_inputs)
        else:
            # If no prompt template or inputs, then specify the prompt as the entire input
            prompt_inputs = prompt

        new_completion_options = {
            **base_completion_options,
            "prompt": rendered_prompt,
        }

        start_time = time.time()
        completion = original_fn(**new_completion_options)

        end_time = time.time()

        create_completion_step_run(
            cls,
            effective_pipeline_slug,
            gentrace_config,
            start_time,
            end_time,
            base_completion_options,
            prompt_template,
            prompt_inputs,
            completion,
            None,
            stream,
            context,
        )

        return completion

    return wrapper


def intercept_completion_async(original_fn, gentrace_config: Configuration):
    @classmethod
    async def wrapper(cls, *args, **kwargs):
        prompt_template = kwargs.get("prompt_template")
        prompt_inputs = kwargs.get("prompt_inputs")
        prompt = kwargs.get("prompt")
        # @deprecated: pipeline_id is deprecated in favor of pipeline_slug
        pipeline_id = kwargs.pop("pipeline_id", None)
        pipeline_slug = kwargs.pop("pipeline_slug", None)

        context = kwargs.pop("gentrace", {})
        stream = kwargs.get("stream")
        base_completion_options = {
            k: v
            for k, v in kwargs.items()
            if k not in ["prompt_template", "prompt_inputs"]
        }

        effective_pipeline_slug = pipeline_slug or pipeline_id

        if stream:
            rendered_prompt = prompt

            if prompt_template and prompt_inputs:
                rendered_prompt = pystache.render(prompt_template, prompt_inputs)
            else:
                # If no prompt template or inputs, then specify the prompt as the entire input
                prompt_inputs = prompt

            new_completion_options = {
                **base_completion_options,
                "prompt": rendered_prompt,
            }

            start_time = time.time()
            completion = await original_fn(**new_completion_options)

            is_self_contained = (
                    not hasattr(cls, "pipeline_run") and effective_pipeline_slug
            )
            if is_self_contained:
                pipeline_run_id = str(uuid.uuid4())

            async def profiled_completion():
                modified_response = []
                async for value in completion:
                    if value and is_self_contained:
                        value["pipelineRunId"] = pipeline_run_id
                    modified_response.append(value)
                    yield value

                end_time = time.time()

                full_response = create_completion_stream_response(modified_response)

                create_completion_step_run(
                    cls,
                    effective_pipeline_slug,
                    gentrace_config,
                    start_time,
                    end_time,
                    base_completion_options,
                    prompt_template,
                    prompt_inputs,
                    full_response,
                    pipeline_run_id if is_self_contained else None,
                    stream,
                    context,
                )

            return profiled_completion()

        rendered_prompt = prompt

        if prompt_template and prompt_inputs:
            rendered_prompt = pystache.render(prompt_template, prompt_inputs)
        else:
            # If no prompt template or inputs, then specify the prompt as the entire input
            prompt_inputs = prompt

        new_completion_options = {
            **base_completion_options,
            "prompt": rendered_prompt,
        }

        start_time = time.time()
        completion = await original_fn(**new_completion_options)
        end_time = time.time()

        create_completion_step_run(
            cls,
            effective_pipeline_slug,
            gentrace_config,
            start_time,
            end_time,
            base_completion_options,
            prompt_template,
            prompt_inputs,
            completion,
            None,
            stream,
            context,
        )
        return completion

    return wrapper


def create_rendered_chat_messages(messages):
    new_messages = []
    for message in messages:
        new_message = copy.deepcopy(message)
        if "contentTemplate" in message and "contentInputs" in message:
            new_message["content"] = pystache.render(
                message["contentTemplate"], message["contentInputs"]
            )

        new_message.pop("contentTemplate", None)
        new_message.pop("contentInputs", None)

        new_messages.append(new_message)

    return new_messages


def intercept_chat_completion(original_fn, gentrace_config: Configuration):
    @classmethod
    def wrapper(cls, *args, **kwargs):
        messages = kwargs.get("messages")
        user = kwargs.get("user")
        # @deprecated: pipeline_id is deprecated in favor of pipeline_slug
        pipeline_id = kwargs.pop("pipeline_id", None)
        pipeline_slug = kwargs.pop("pipeline_slug", None)
        stream = kwargs.get("stream")
        context = kwargs.pop("gentrace", {})

        model_params = {
            k: v for k, v in kwargs.items() if k not in ["messages", "user"]
        }

        effective_pipeline_slug = pipeline_slug or pipeline_id

        content_templates_array = [
            message["contentTemplate"] if "contentTemplate" in message else None
            for message in messages
        ]

        content_inputs_array = [
            message["contentInputs"] if "contentInputs" in message else None
            for message in messages
        ]

        if stream:
            rendered_messages = create_rendered_chat_messages(messages)
            new_kwargs = dict(kwargs, messages=rendered_messages)
            start_time = time.time()
            completion = original_fn(**new_kwargs)

            is_self_contained = (
                    not hasattr(cls, "pipeline_run") and effective_pipeline_slug
            )
            if is_self_contained:
                pipeline_run_id = str(uuid.uuid4())

            def profiled_completion():
                modified_response = []
                for value in completion:
                    if value and is_self_contained:
                        value["pipelineRunId"] = pipeline_run_id
                    modified_response.append(value)
                    yield value

                end_time = time.time()

                elapsed_time = int((end_time - start_time) * 1000)

                pipeline_run = (
                    cls.pipeline_run if hasattr(cls, "pipeline_run") else None
                )

                full_response = create_chat_completion_stream_response(
                    modified_response
                )
                if is_self_contained:
                    pipeline = Pipeline(
                        slug=effective_pipeline_slug,
                        api_key=gentrace_config.access_token,
                        host=gentrace_config.host,
                    )

                    pipeline_run = PipelineRun(
                        pipeline=pipeline, id=pipeline_run_id, context=context
                    )

                if pipeline_run:
                    pipeline_run.add_step_run(
                        OpenAICreateChatCompletionStepRun(
                            elapsed_time,
                            to_date_string(start_time),
                            to_date_string(end_time),
                            {
                                "messages": messages,
                                "user": user,
                                "contentInputs": content_inputs_array,
                            },
                            {
                                **model_params,
                                "contentTemplates": content_templates_array,
                            },
                            full_response,
                            context,
                        )
                    )

                    if is_self_contained:
                        pipeline_run.submit()

            return profiled_completion()

        rendered_messages = create_rendered_chat_messages(messages)

        new_kwargs = dict(kwargs, messages=rendered_messages)

        start_time = time.time()
        completion = original_fn(**new_kwargs)
        end_time = time.time()

        elapsed_time = int((end_time - start_time) * 1000)

        pipeline_run = cls.pipeline_run if hasattr(cls, "pipeline_run") else None

        is_self_contained = not pipeline_run and effective_pipeline_slug
        if is_self_contained:
            pipeline = Pipeline(
                slug=effective_pipeline_slug,
                api_key=gentrace_config.access_token,
                host=gentrace_config.host,
            )

            pipeline_run = PipelineRun(
                pipeline=pipeline,
                context=context,
            )

        if pipeline_run:
            pipeline_run.add_step_run(
                OpenAICreateChatCompletionStepRun(
                    elapsed_time,
                    to_date_string(start_time),
                    to_date_string(end_time),
                    {
                        "messages": messages,
                        "user": user,
                        "contentInputs": content_inputs_array,
                    },
                    {**model_params, "contentTemplates": content_templates_array},
                    completion,
                    context,
                )
            )

            if is_self_contained:
                submit_result = pipeline_run.submit()
                completion["pipelineRunId"] = (
                    submit_result["pipelineRunId"]
                    if "pipelineRunId" in submit_result
                    else None
                )

        return completion

    return wrapper


def intercept_chat_completion_async(original_fn, gentrace_config: Configuration):
    @classmethod
    async def wrapper(cls, *args, **kwargs):
        messages = kwargs.get("messages")
        user = kwargs.get("user")
        stream = kwargs.get("stream")
        # @deprecated: pipeline_id is deprecated in favor of pipeline_slug
        pipeline_id = kwargs.pop("pipeline_id", None)
        pipeline_slug = kwargs.pop("pipeline_slug", None)
        context = kwargs.pop("gentrace", {})

        model_params = {
            k: v for k, v in kwargs.items() if k not in ["messages", "user"]
        }

        content_templates_array = [
            message["contentTemplate"] if "contentTemplate" in message else None
            for message in messages
        ]

        content_inputs_array = [
            message["contentInputs"] if "contentInputs" in message else None
            for message in messages
        ]

        effective_pipeline_slug = pipeline_slug or pipeline_id

        rendered_messages = create_rendered_chat_messages(messages)
        new_kwargs = dict(kwargs, messages=rendered_messages)

        start_time = time.time()
        completion = await original_fn(**new_kwargs)

        if stream:
            is_self_contained = (
                    not hasattr(cls, "pipeline_run") and effective_pipeline_slug
            )
            if is_self_contained:
                pipeline_run_id = str(uuid.uuid4())

            async def profiled_completion():
                modified_response = []
                async for value in completion:
                    if value and is_self_contained:
                        value["pipelineRunId"] = pipeline_run_id
                    modified_response.append(value)
                    yield value

                end_time = time.time()

                full_response = create_chat_completion_stream_response(
                    modified_response
                )

                elapsed_time = int((end_time - start_time) * 1000)

                pipeline_run = (
                    cls.pipeline_run if hasattr(cls, "pipeline_run") else None
                )

                if is_self_contained:
                    pipeline = Pipeline(
                        slug=effective_pipeline_slug,
                        api_key=gentrace_config.access_token,
                        host=gentrace_config.host,
                    )

                    pipeline_run = PipelineRun(
                        pipeline=pipeline,
                        id=pipeline_run_id,
                        context=context,
                    )

                if pipeline_run:
                    pipeline_run.add_step_run(
                        OpenAICreateChatCompletionStepRun(
                            elapsed_time,
                            to_date_string(start_time),
                            to_date_string(end_time),
                            {
                                "messages": messages,
                                "user": user,
                                "contentInputs": content_inputs_array,
                            },
                            {
                                **model_params,
                                "contentTemplates": content_templates_array,
                            },
                            full_response,
                            context,
                        )
                    )

                    if is_self_contained:
                        pipeline_run.submit()

            return profiled_completion()

        end_time = time.time()

        elapsed_time = int((end_time - start_time) * 1000)

        pipeline_run = cls.pipeline_run if hasattr(cls, "pipeline_run") else None

        is_self_contained = not pipeline_run and effective_pipeline_slug

        if is_self_contained:
            pipeline = Pipeline(
                slug=effective_pipeline_slug,
                api_key=gentrace_config.access_token,
                host=gentrace_config.host,
            )

            pipeline_run = PipelineRun(
                pipeline=pipeline,
                context=context,
            )

        if pipeline_run:
            pipeline_run.add_step_run(
                OpenAICreateChatCompletionStepRun(
                    elapsed_time,
                    to_date_string(start_time),
                    to_date_string(end_time),
                    {
                        "messages": messages,
                        "user": user,
                        "contentInputs": content_inputs_array,
                    },
                    {
                        **model_params,
                        "contentTemplates": content_templates_array,
                    },
                    completion,
                    context,
                )
            )

            if is_self_contained:
                submit_result = pipeline_run.submit()
                completion["pipelineRunId"] = (
                    submit_result["pipelineRunId"]
                    if "pipelineRunId" in submit_result
                    else None
                )
        return completion

    return wrapper


def intercept_embedding(original_fn, gentrace_config: Configuration):
    @classmethod
    def wrapper(cls, *args, **kwargs):
        model = kwargs.get("model")
        input_params = {k: v for k, v in kwargs.items() if k not in ["model"]}
        pipeline_id = kwargs.pop("pipeline_id", None)
        pipeline_slug = kwargs.pop("pipeline_slug", None)
        context = kwargs.pop("gentrace", {})
        input_params = {k: v for k, v in kwargs.items() if k not in ["model"]}

        effective_pipeline_slug = pipeline_slug or pipeline_id

        start_time = time.time()
        completion = original_fn(**kwargs)
        end_time = time.time()

        elapsed_time = int((end_time - start_time) * 1000)

        pipeline_run = cls.pipeline_run if hasattr(cls, "pipeline_run") else None

        is_self_contained = not pipeline_run and effective_pipeline_slug

        if is_self_contained:
            pipeline = Pipeline(
                slug=effective_pipeline_slug,
                api_key=gentrace_config.access_token,
                host=gentrace_config.host,
            )

            pipeline_run = PipelineRun(
                pipeline=pipeline,
                context=context,
            )

        if pipeline_run:
            pipeline_run.add_step_run(
                OpenAICreateEmbeddingStepRun(
                    elapsed_time,
                    to_date_string(start_time),
                    to_date_string(end_time),
                    input_params,
                    {"model": model},
                    completion,
                    context,
                )
            )

            if is_self_contained:
                submit_result = pipeline_run.submit()
                completion["pipelineRunId"] = (
                    submit_result["pipelineRunId"]
                    if "pipelineRunId" in submit_result
                    else None
                )
        return completion

    return wrapper


def intercept_embedding_async(original_fn, gentrace_config: Configuration):
    @classmethod
    async def wrapper(cls, *args, **kwargs):
        model = kwargs.get("model")
        # @deprecated: pipeline_id is deprecated, use pipeline_slug instead
        pipeline_id = kwargs.pop("pipeline_id", None)
        pipeline_slug = kwargs.pop("pipeline_slug", None)

        context = kwargs.pop("gentrace", {})
        input_params = {k: v for k, v in kwargs.items() if k not in ["model"]}

        effective_pipeline_slug = pipeline_slug or pipeline_id

        start_time = time.time()
        completion = await original_fn(**kwargs)
        end_time = time.time()

        elapsed_time = int((end_time - start_time) * 1000)

        pipeline_run = cls.pipeline_run if hasattr(cls, "pipeline_run") else None

        is_self_contained = not pipeline_run and effective_pipeline_slug

        if is_self_contained:
            pipeline = Pipeline(
                slug=effective_pipeline_slug,
                api_key=gentrace_config.access_token,
                host=gentrace_config.host,
            )

            pipeline_run = PipelineRun(pipeline=pipeline, context=context)

        if pipeline_run:
            pipeline_run.add_step_run(
                OpenAICreateEmbeddingStepRun(
                    elapsed_time,
                    to_date_string(start_time),
                    to_date_string(end_time),
                    input_params,
                    {"model": model},
                    completion,
                    context,
                )
            )

            if is_self_contained:
                submit_result = pipeline_run.submit()
                completion["pipelineRunId"] = (
                    submit_result["pipelineRunId"]
                    if "pipelineRunId" in submit_result
                    else None
                )

        return completion

    return wrapper


def swap_methods(cls, attribute: str, gentrace_config: Configuration, intercept_fn):
    has_old_sync_saved = hasattr(cls, attribute + "_old")
    original_create = getattr(cls, attribute)
    if has_old_sync_saved:
        original_create = getattr(cls, attribute + "_old")
    else:
        setattr(cls, attribute + "_old", getattr(cls, attribute))

    setattr(cls, attribute, intercept_fn(original_create, gentrace_config))


def annotate_openai_module(
        gentrace_config: Configuration,
):
    import openai

    for name, cls in vars(openai.api_resources).items():
        if isinstance(cls, type):
            if name == "Completion":
                swap_methods(cls, "create", gentrace_config, intercept_completion)
                swap_methods(
                    cls, "acreate", gentrace_config, intercept_completion_async
                )
            elif name == "ChatCompletion":
                swap_methods(cls, "create", gentrace_config, intercept_chat_completion)
                swap_methods(
                    cls, "acreate", gentrace_config, intercept_chat_completion_async
                )

            elif name == "Embedding":
                swap_methods(cls, "create", gentrace_config, intercept_embedding)
                swap_methods(cls, "acreate", gentrace_config, intercept_embedding_async)

            setattr(openai.api_resources, name, cls)


def annotate_pipeline_handler(
        handler: OpenAIPipelineHandler,
        gentrace_config: Configuration,
        pipeline_run: Optional[PipelineRun] = None,
):
    import openai

    for name, cls in vars(openai.api_resources).items():
        if isinstance(cls, type):
            # Create new class that inherits from the original class, don't directly monkey patch
            # the original class
            new_class = type(name, (cls,), {})
            if name == "Completion":
                original_create = (
                    cls.create_old if hasattr(cls, "create_old") else cls.create
                )
                new_class.create = intercept_completion(
                    original_create, gentrace_config
                )
                original_acreate = (
                    cls.acreate_old if hasattr(cls, "acreate_old") else cls.acreate
                )
                new_class.acreate = intercept_completion_async(
                    original_acreate, gentrace_config
                )
            elif name == "ChatCompletion":
                original_create = (
                    cls.create_old if hasattr(cls, "create_old") else cls.create
                )
                new_class.create = intercept_chat_completion(
                    original_create, gentrace_config
                )
                original_acreate = (
                    cls.acreate_old if hasattr(cls, "acreate_old") else cls.acreate
                )
                new_class.acreate = intercept_chat_completion_async(
                    original_acreate, gentrace_config
                )
            elif name == "Embedding":
                original_create = (
                    cls.create_old if hasattr(cls, "create_old") else cls.create
                )
                new_class.create = intercept_embedding(original_create, gentrace_config)
                original_acreate = (
                    cls.acreate_old if hasattr(cls, "acreate_old") else cls.acreate
                )
                new_class.acreate = intercept_embedding_async(
                    original_acreate, gentrace_config
                )

            new_class.pipeline_run = pipeline_run
            setattr(handler, name, new_class)

    return handler


class OpenAICreateCompletionStepRun(StepRun):
    def __init__(
            self,
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
            error,
    ):
        super().__init__(
            "openai",
            "openai_createCompletion",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
            error,
        )
        self.response = response


class OpenAICreateChatCompletionStepRun(StepRun):
    def __init__(
            self,
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
            error,
    ):
        super().__init__(
            "openai",
            "openai_createChatCompletion",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
            error,
        )
        self.response = response


class OpenAICreateEmbeddingStepRun(StepRun):
    def __init__(
            self,
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
            error,
    ):
        super().__init__(
            "openai",
            "openai_createEmbedding",
            elapsed_time,
            start_time,
            end_time,
            inputs,
            model_params,
            response,
            context,
            error,
        )
        self.response = response



# coding: utf-8

# flake8: noqa

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.27.0
    Generated by: https://openapi-generator.tech
"""

__version__ = "1.0.0"

# import ApiClient
from gentrace.api_client import ApiClient

# import Configuration
from gentrace.configuration import Configuration

from gentrace.providers.utils import is_openai_v1

# import exceptions
from gentrace.exceptions import OpenApiException
from gentrace.exceptions import ApiAttributeError
from gentrace.exceptions import ApiTypeError
from gentrace.exceptions import ApiValueError
from gentrace.exceptions import ApiKeyError
from gentrace.exceptions import ApiException

from gentrace.providers.evaluation import *
from gentrace.providers.pipeline import Pipeline
from gentrace.providers.pipeline_run import PipelineRun, flush, get_run
from gentrace.providers.step_run import StepRun
from gentrace.providers.utils import to_date_string
from gentrace.providers.files import *

from gentrace.providers.getters import *
from gentrace.providers.init import init, deinit

if is_openai_v1():
    from gentrace.providers.llms.handlers.openai_v1.simple import SimpleGentraceSyncOpenAITyped as OpenAI
    from gentrace.providers.llms.handlers.openai_v1.simple import SimpleGentraceAsyncOpenAITyped as AsyncOpenAI



# @deprecated: use gentrace.providers.init.init() instead to set the Gentrace
# API key
api_key = ""

# @deprecated: use gentrace.providers.init.init() instead to set the Gentrace
# host
host = ""

# @deprecated: use gentrace.providers.init.init() instead to set the Gentrace
# log level
log_level = "warn"


# coding: utf-8

"""
    Gentrace API

    These API routes are designed to ingest events from clients.  # noqa: E501

    The version of the OpenAPI document: 0.27.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from gentrace import schemas  # noqa: F401


class CreateMultipleTestCases(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "testCases",
        }
        
        class properties:
            
            
            class testCases(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    
                    
                    class items(
                        schemas.DictSchema
                    ):
                    
                    
                        class MetaOapg:
                            required = {
                                "inputs",
                                "name",
                            }
                            
                            class properties:
                                
                                
                                class name(
                                    schemas.StrSchema
                                ):
                                    pass
                                
                                
                                class inputs(
                                    schemas.DictSchema
                                ):
                                
                                
                                    class MetaOapg:
                                        additional_properties = schemas.AnyTypeSchema
                                    
                                    def __getitem__(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
                                        # dict_instance[name] accessor
                                        return super().__getitem__(name)
                                    
                                    def get_item_oapg(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
                                        return super().get_item_oapg(name)
                                
                                    def __new__(
                                        cls,
                                        *_args: typing.Union[dict, frozendict.frozendict, ],
                                        _configuration: typing.Optional[schemas.Configuration] = None,
                                        **kwargs: typing.Union[MetaOapg.additional_properties, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                                    ) -> 'inputs':
                                        return super().__new__(
                                            cls,
                                            *_args,
                                            _configuration=_configuration,
                                            **kwargs,
                                        )
                                expectedOutputs = schemas.DictSchema
                                __annotations__ = {
                                    "name": name,
                                    "inputs": inputs,
                                    "expectedOutputs": expectedOutputs,
                                }
                        
                        inputs: MetaOapg.properties.inputs
                        name: MetaOapg.properties.name
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["name"]) -> MetaOapg.properties.name: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["inputs"]) -> MetaOapg.properties.inputs: ...
                        
                        @typing.overload
                        def __getitem__(self, name: typing_extensions.Literal["expectedOutputs"]) -> MetaOapg.properties.expectedOutputs: ...
                        
                        @typing.overload
                        def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
                        
                        def __getitem__(self, name: typing.Union[typing_extensions.Literal["name", "inputs", "expectedOutputs", ], str]):
                            # dict_instance[name] accessor
                            return super().__getitem__(name)
                        
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["name"]) -> MetaOapg.properties.name: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["inputs"]) -> MetaOapg.properties.inputs: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: typing_extensions.Literal["expectedOutputs"]) -> typing.Union[MetaOapg.properties.expectedOutputs, schemas.Unset]: ...
                        
                        @typing.overload
                        def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
                        
                        def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["name", "inputs", "expectedOutputs", ], str]):
                            return super().get_item_oapg(name)
                        
                    
                        def __new__(
                            cls,
                            *_args: typing.Union[dict, frozendict.frozendict, ],
                            inputs: typing.Union[MetaOapg.properties.inputs, dict, frozendict.frozendict, ],
                            name: typing.Union[MetaOapg.properties.name, str, ],
                            expectedOutputs: typing.Union[MetaOapg.properties.expectedOutputs, dict, frozendict.frozendict, schemas.Unset] = schemas.unset,
                            _configuration: typing.Optional[schemas.Configuration] = None,
                            **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                        ) -> 'items':
                            return super().__new__(
                                cls,
                                *_args,
                                inputs=inputs,
                                name=name,
                                expectedOutputs=expectedOutputs,
                                _configuration=_configuration,
                                **kwargs,
                            )
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[MetaOapg.items, dict, frozendict.frozendict, ]], typing.List[typing.Union[MetaOapg.items, dict, frozendict.frozendict, ]]],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'testCases':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)
            
            
            class pipelineSlug(
                schemas.StrSchema
            ):
                pass
            
            
            class datasetId(
                schemas.StrSchema
            ):
                pass
            __annotations__ = {
                "testCases": testCases,
                "pipelineSlug": pipelineSlug,
                "datasetId": datasetId,
            }
    
    testCases: MetaOapg.properties.testCases
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["testCases"]) -> MetaOapg.properties.testCases: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["pipelineSlug"]) -> MetaOapg.properties.pipelineSlug: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["datasetId"]) -> MetaOapg.properties.datasetId: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["testCases", "pipelineSlug", "datasetId", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["testCases"]) -> MetaOapg.properties.testCases: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["pipelineSlug"]) -> typing.Union[MetaOapg.properties.pipelineSlug, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["datasetId"]) -> typing.Union[MetaOapg.properties.datasetId, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["testCases", "pipelineSlug", "datasetId", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        testCases: typing.Union[MetaOapg.properties.testCases, list, tuple, ],
        pipelineSlug: typing.Union[MetaOapg.properties.pipelineSlug, str, schemas.Unset] = schemas.unset,
        datasetId: typing.Union[MetaOapg.properties.datasetId, str, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'CreateMultipleTestCases':
        return super().__new__(
            cls,
            *_args,
            testCases=testCases,
            pipelineSlug=pipelineSlug,
            datasetId=datasetId,
            _configuration=_configuration,
            **kwargs,
        )
